"""
The following utility functions are used for data visualization
https://plotly.com/python/reference/
"""
#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from typing import TYPE_CHECKING
from typing import Optional
from typing import Union
from typing import Any
import sys
import plotly.graph_objects as go  # type: ignore
import plotly.io as pio
import numpy as np
import numpy.typing as npt
from . import graphics_common, graphics_common_3d
from ..ops import element
from ..ops import node as node_module
from ..transformations import local_axes_from_points_and_angle

if TYPE_CHECKING:
    from ..model import Model
    from ..load_case import LoadCase

nparr = npt.NDArray[np.float64]


def add_data__nodes(data_dict, mdl, load_case):
    """
    Adds a trace containing nodes

    Arguments:
      data_dict:
        list of dictionaries containing figure data
      mdl: the model to be visualized
      load_case: the load_case to be visualized

    """

    # TODO: this was written very sloppily, because I was in a rush.
    # I need to improve this code, avoid code repetition
    # TODO: merge this method with the other methods that plot nodes.
    list_of_nodes = mdl.list_of_primary_nodes()
    x_list = [node.coords[0] for node in list_of_nodes]
    y_list = [node.coords[1] for node in list_of_nodes]
    z_list = [node.coords[2] for node in list_of_nodes]
    customdata_lst = []
    restraints = [node.restraint for node in list_of_nodes]
    restraint_symbols = []
    for node in list_of_nodes:
        if True in node.restraint:
            restraint_symbols.append("fixed")
        else:
            restraint_symbols.append("free")
    for node in list_of_nodes:
        if load_case:
            customdata_lst.append(
                (
                    node.uid,
                    *load_case.node_mass[node.uid].val,
                    *load_case.node_loads[node.uid].val,
                )
            )
        else:
            customdata_lst.append((node.uid,))
    customdata: nparr = np.array(customdata_lst, dtype="object")
    if load_case:
        data_dict.append(
            {
                "name": "Primary nodes",
                "type": "scatter3d",
                "mode": "markers",
                "x": x_list,
                "y": y_list,
                "z": z_list,
                "customdata": customdata,
                "text": restraints,
                "hovertemplate": (
                    "Coordinates: (%{x:.2f}, %{y:.2f}, %{z:.2f})<br>"
                    + "Restraint: %{text}<br>"
                    + "Mass: (%{customdata[1]:.3g}, "
                    + "%{customdata[2]:.3g}, %{customdata[3]:.3g}, "
                    + "%{customdata[4]:.3g}, %{customdata[5]:.3g}, "
                    + "%{customdata[6]:.3g})<br>"
                    + "Load: (%{customdata[7]:.3g}, "
                    + "%{customdata[8]:.3g}, %{customdata[9]:.3g}, "
                    + "%{customdata[10]:.3g}, %{customdata[11]:.3g}, "
                    + "%{customdata[12]:.3g})"
                    + "<extra>Node: %{customdata[0]:d}</extra>"
                ),
                "marker": {
                    "symbol": [
                        graphics_common_3d.node_marker[sym][0]
                        for sym in restraint_symbols
                    ],
                    "color": graphics_common.NODE_PRIMARY_COLOR,
                    "size": [
                        graphics_common_3d.node_marker[sym][1]
                        for sym in restraint_symbols
                    ],
                    "line": {
                        "color": graphics_common.NODE_PRIMARY_COLOR,
                        "width": 4,
                    },
                },
            }
        )
    else:
        data_dict.append(
            {
                "name": "Primary nodes",
                "type": "scatter3d",
                "mode": "markers",
                "x": x_list,
                "y": y_list,
                "z": z_list,
                "customdata": customdata,
                "text": restraints,
                "hovertemplate": (
                    "Coordinates: (%{x:.2f}, %{y:.2f}, %{z:.2f})<br>"
                    + "Restraint: %{text}<br>"
                    + "<extra>Node: %{customdata[0]:d}</extra>"
                ),
                "marker": {
                    "symbol": [
                        graphics_common_3d.node_marker[sym][0]
                        for sym in restraint_symbols
                    ],
                    "color": graphics_common.NODE_PRIMARY_COLOR,
                    "size": [
                        graphics_common_3d.node_marker[sym][1]
                        for sym in restraint_symbols
                    ],
                    "line": {
                        "color": graphics_common.NODE_PRIMARY_COLOR,
                        "width": 4,
                    },
                },
            }
        )


def add_data__parent_nodes(
    data_dict: list[dict[str, object]], load_case: LoadCase
) -> None:
    """
    Adds a trace containing parent nodes

    Arguments:
      data_dict:
        list of dictionaries containing figure data
      mdl: the model to be visualized
      load_case: the load_case to be visualized

    """

    list_of_nodes = load_case.parent_nodes.values()
    x_list = [node.coords[0] for node in list_of_nodes]
    y_list = [node.coords[1] for node in list_of_nodes]
    z_list = [node.coords[2] for node in list_of_nodes]
    customdata_list = []
    restraints = [node.restraint for node in list_of_nodes]
    for node in list_of_nodes:
        customdata_list.append(
            (
                node.uid,
                *load_case.node_mass[node.uid].val,
                *load_case.node_loads[node.uid].val,
            )
        )

    customdata: nparr = np.array(customdata_list, dtype="object")
    data_dict.append(
        {
            "type": "scatter3d",
            "mode": "markers",
            "x": x_list,
            "y": y_list,
            "z": z_list,
            "customdata": customdata,
            "text": restraints,
            "hovertemplate": (
                "Coordinates: (%{x:.2f}, %{y:.2f}, %{z:.2f})<br>"
                + "Restraint: %{text}<br>"
                + "Mass: (%{customdata[1]:.3g}, "
                + "%{customdata[2]:.3g}, %{customdata[3]:.3g}, "
                + "%{customdata[4]:.3g}, %{customdata[5]:.3g}, "
                + "%{customdata[6]:.3g})<br>"
                + "Load: (%{customdata[7]:.3g}, "
                + "%{customdata[8]:.3g}, %{customdata[9]:.3g}, "
                + "%{customdata[10]:.3g}, %{customdata[11]:.3g}, "
                + "%{customdata[12]:.3g})"
                + "<extra>Parent Node: %{customdata[0]:d}</extra>"
            ),
            "marker": {
                "symbol": [graphics_common_3d.node_marker["parent"][0]]
                * len(list_of_nodes),
                "color": graphics_common.NODE_PRIMARY_COLOR,
                "size": [graphics_common_3d.node_marker["parent"][1]]
                * len(list_of_nodes),
                "line": {
                    "color": graphics_common.NODE_PRIMARY_COLOR,
                    "width": 4,
                },
            },
        }
    )


def add_data__internal_nodes(
    data_dict: list[dict[str, object]],
    mdl: Model,
    load_case: Optional[LoadCase] = None,
) -> None:
    """
    Adds a trace containing internal nodes

    Arguments:
      data_dict:
        list of dictionaries containing figure data
      mdl: the model to be visualized
      load_case: the load_case to be visualized

    """

    list_of_nodes = mdl.list_of_internal_nodes()
    x_list = [node.coords[0] for node in list_of_nodes]
    y_list = [node.coords[1] for node in list_of_nodes]
    z_list = [node.coords[2] for node in list_of_nodes]
    customdata = []
    restraints = [node.restraint for node in list_of_nodes]
    restraint_symbols = []
    for node in list_of_nodes:
        if True in node.restraint:
            restraint_symbols.append("fixed")
        elif node.visibility.connected_to_zerolength:
            restraint_symbols.append("release")
        else:
            restraint_symbols.append("internal")
    for node in list_of_nodes:
        if load_case:
            customdata.append(
                (
                    node.uid,
                    *load_case.node_mass[node.uid].val,
                    *load_case.node_loads[node.uid].val,
                )
            )
        else:
            customdata.append((node.uid,))
    if load_case:
        data_dict.append(
            {
                "name": "Internal nodes",
                "type": "scatter3d",
                "mode": "markers",
                "x": x_list,
                "y": y_list,
                "z": z_list,
                "customdata": customdata,
                "text": restraints,
                "hovertemplate": (
                    "Coordinates: (%{x:.2f}, %{y:.2f}, %{z:.2f})<br>"
                    + "Restraint: %{text}<br>"
                    + "Mass: (%{customdata[1]:.3g}, "
                    + "%{customdata[2]:.3g}, %{customdata[3]:.3g})<br>"
                    + "Load: (%{customdata[4]:.3g}, "
                    + "%{customdata[5]:.3g}, %{customdata[6]:.3g})"
                    + "<extra>Node: %{customdata[0]:d}</extra>"
                ),
                "marker": {
                    "symbol": [
                        graphics_common_3d.node_marker[sym][0]
                        for sym in restraint_symbols
                    ],
                    "color": graphics_common.NODE_INTERNAL_COLOR,
                    "size": [
                        graphics_common_3d.node_marker[sym][1]
                        for sym in restraint_symbols
                    ],
                    "line": {
                        "color": graphics_common.NODE_INTERNAL_COLOR,
                        "width": 2,
                    },
                },
            }
        )
    else:
        data_dict.append(
            {
                "name": "Internal nodes",
                "type": "scatter3d",
                "mode": "markers",
                "x": x_list,
                "y": y_list,
                "z": z_list,
                "customdata": customdata,
                "text": restraints,
                "hovertemplate": (
                    "Coordinates: (%{x:.2f}, %{y:.2f}, %{z:.2f})<br>"
                    + "Restraint: %{text}<br>"
                    + "<extra>Node: %{customdata[0]:d}</extra>"
                ),
                "marker": {
                    "symbol": [
                        graphics_common_3d.node_marker[sym][0]
                        for sym in restraint_symbols
                    ],
                    "color": graphics_common.NODE_INTERNAL_COLOR,
                    "size": [
                        graphics_common_3d.node_marker[sym][1]
                        for sym in restraint_symbols
                    ],
                    "line": {
                        "color": graphics_common.NODE_INTERNAL_COLOR,
                        "width": 2,
                    },
                },
            }
        )


def add_data__release_nodes(
    data_dict: list[dict[str, object]], list_of_nodes: list[node_module.Node]
) -> None:
    """
    Adds a trace containing release nodes

    Arguments:
      data_dict:
        list of dictionaries containing figure data
      list_of_nodes: List of nodes where a release is present.
      mdl: the model to be visualized
      load_case: the load_case to be visualized

    """

    x_list = [node.coords[0] for node in list_of_nodes]
    y_list = [node.coords[1] for node in list_of_nodes]
    z_list = [node.coords[2] for node in list_of_nodes]
    data_dict.append(
        {
            "type": "scatter3d",
            "mode": "markers",
            "x": x_list,
            "y": y_list,
            "z": z_list,
            "hoverinfo": "skip",
            "marker": {
                "symbol": graphics_common_3d.node_marker["pinned"][0],
                "color": graphics_common.NODE_INTERNAL_COLOR,
                "size": graphics_common_3d.node_marker["pinned"][1],
                "line": {
                    "color": graphics_common.NODE_INTERNAL_COLOR,
                    "width": 2,
                },
            },
        }
    )


def add_data__frames(
    data_dict: list[dict[str, object]],
    mdl: Model,
    load_case: Optional[LoadCase] = None,
) -> None:
    """
    Adds a trace containing frame element centroidal axis lines

    Arguments:
      data_dict:
        list of dictionaries containing figure data
      mdl: the model to be visualized
      load_case: the load_case to be visualized

    """

    line_elems: list[
        Union[element.ElasticBeamColumn, element.DispBeamColumn]
    ] = []

    # >>>
    # it feels ridiculous to do this, but it's the only way I am
    # getting rid of type checking warnings.
    elms_elastic = [
        elm
        for elm in mdl.list_of_specific_element(element.ElasticBeamColumn)
        if isinstance(elm, element.ElasticBeamColumn)
    ]
    line_elems.extend(elms_elastic)
    elms_disp = [
        elm
        for elm in mdl.list_of_specific_element(element.DispBeamColumn)
        if isinstance(elm, element.DispBeamColumn)
    ]
    line_elems.extend(elms_disp)
    # <<<

    if not line_elems:
        return

    x_list: list[Optional[float]] = []
    y_list: list[Optional[float]] = []
    z_list: list[Optional[float]] = []
    customdata_list = []
    section_names = []
    for elm in line_elems:
        if elm.visibility.hidden_at_line_plots:
            continue
        p_i = np.array(elm.nodes[0].coords) + elm.geomtransf.offset_i
        p_j = np.array(elm.nodes[1].coords) + elm.geomtransf.offset_j
        section_name = elm.section.name
        section_names.extend([section_name] * 3)
        x_list.extend((p_i[0], p_j[0], None))
        y_list.extend((p_i[1], p_j[1], None))
        z_list.extend((p_i[2], p_j[2], None))
        if load_case:
            customdata_list.append(
                (
                    elm.uid,
                    *load_case.line_element_udl[elm.uid].val,
                    elm.nodes[0].uid,
                    elm.parent_component.uid,
                )
            )
            customdata_list.append(
                (
                    elm.uid,
                    *load_case.line_element_udl[elm.uid].val,
                    elm.nodes[1].uid,
                    elm.parent_component.uid,
                )
            )
            customdata_list.append((None,) * 6)
        else:
            customdata_list.append(
                (elm.uid, elm.nodes[0].uid, elm.parent_component.uid)
            )
            customdata_list.append(
                (elm.uid, elm.nodes[1].uid, elm.parent_component.uid)
            )
            customdata_list.append((None,) * 3)

    if load_case:
        customdata: nparr = np.array(customdata_list, dtype="object")
        data_dict.append(
            {
                "name": "Frame elements",
                "type": "scatter3d",
                "mode": "lines",
                "x": x_list,
                "y": y_list,
                "z": z_list,
                "text": section_names,
                "customdata": customdata,
                "hovertemplate": (
                    "Section: %{text}<br>"
                    + "UDL (local): (%{customdata[1]:.3g}, "
                    + "%{customdata[2]:.3g}, %{customdata[3]:.3g})"
                    + "<extra>Element: %{customdata[0]:d}<br>"
                    + "Node @ this end: %{customdata[4]:d}<br>"
                    "Parent: %{customdata[5]}</extra>"
                ),
                "line": {"width": 5, "color": graphics_common.FRAME_COLOR},
            }
        )
    else:
        customdata = np.array(customdata_list, dtype="object")
        data_dict.append(
            {
                "name": "Frame elements",
                "type": "scatter3d",
                "mode": "lines",
                "x": x_list,
                "y": y_list,
                "z": z_list,
                "text": section_names,
                "customdata": customdata,
                "hovertemplate": (
                    "Section: %{text}<br>"
                    + "<extra>Element: %{customdata[0]:d}<br>"
                    + "Node @ this end: %{customdata[1]:d}<br>"
                    "Parent: %{customdata[2]}</extra>"
                ),
                "line": {"width": 5, "color": graphics_common.FRAME_COLOR},
            }
        )


def add_data__bars(
    data_dict: list[dict[str, object]],
    mdl: Model,
    load_case: Optional[LoadCase] = None,
) -> None:
    """
    Adds a trace containing frame element centroidal axis lines

    Arguments:
      data_dict: dictionary containing figure data
      mdl: the model to be visualized
      load_case: the load_case to be visualized

    """

    line_elems = [
        elm
        for elm in mdl.list_of_specific_element(element.TrussBar)
        if isinstance(elm, element.TrussBar)
    ]
    if not line_elems:
        return
    x_list: list[Optional[float]] = []
    y_list: list[Optional[float]] = []
    z_list: list[Optional[float]] = []
    customdata_list: list[tuple[Any, ...]] = []
    section_areas = []
    for elm in line_elems:
        if elm.visibility.hidden_at_line_plots:
            continue
        p_i = np.array(elm.nodes[0].coords)
        p_j = np.array(elm.nodes[1].coords)
        section_area = elm.area

        section_areas.extend([section_area] * 3)
        x_list.extend((p_i[0], p_j[0], None))
        y_list.extend((p_i[1], p_j[1], None))
        z_list.extend((p_i[2], p_j[2], None))
        if load_case:
            customdata_list.append(
                (
                    elm.uid,
                    elm.nodes[0].uid,
                    elm.parent_component.uid,
                )
            )
            customdata_list.append(
                (
                    elm.uid,
                    elm.nodes[1].uid,
                    elm.parent_component.uid,
                )
            )
            customdata_list.append((None,) * 6)
        else:
            customdata_list.append(
                (elm.uid, elm.nodes[0].uid, elm.parent_component.uid)
            )
            customdata_list.append(
                (elm.uid, elm.nodes[1].uid, elm.parent_component.uid)
            )
            customdata_list.append((None,) * 3)

    if load_case:
        customdata: nparr = np.array(customdata_list, dtype="object")
        data_dict.append(
            {
                "name": "Truss elements",
                "type": "scatter3d",
                "mode": "lines",
                "x": x_list,
                "y": y_list,
                "z": z_list,
                "text": section_areas,
                "customdata": customdata,
                "hovertemplate": (
                    "Section area: %{text}<br>"
                    + "<extra>Element: %{customdata[0]:d}<br>"
                    + "Node @ this end: %{customdata[1]:d}<br>"
                    "Parent: %{customdata[2]}</extra>"
                ),
                "line": {"width": 5, "color": graphics_common.FRAME_COLOR},
            }
        )
    else:
        customdata = np.array(customdata_list, dtype="object")
        data_dict.append(
            {
                "name": "Truss elements",
                "type": "scatter3d",
                "mode": "lines",
                "x": x_list,
                "y": y_list,
                "z": z_list,
                "text": section_areas,
                "customdata": customdata,
                "hovertemplate": (
                    "Section area: %{text}<br>"
                    + "<extra>Element: %{customdata[0]:d}<br>"
                    + "Node @ this end: %{customdata[1]:d}<br>"
                    "Parent: %{customdata[2]}</extra>"
                ),
                "line": {"width": 5, "color": graphics_common.FRAME_COLOR},
            }
        )


def add_data__twonodelinks(
    data_dict: list[dict[str, object]], mdl: Model
) -> None:
    """
    Adds a trace containing twonodelink elements

    Arguments:
      data_dict: dictionary containing figure data
      mdl: the model to be visualized

    """

    link_elems = [
        elm
        for elm in mdl.list_of_elements()
        if isinstance(elm, element.TwoNodeLink)
    ]

    if not link_elems:
        return
    x_list: list[Optional[float]] = []
    y_list: list[Optional[float]] = []
    z_list: list[Optional[float]] = []
    customdata_list: list[tuple[Any, ...]] = []
    for elm in link_elems:
        p_i: nparr = np.array(elm.nodes[0].coords)
        p_j: nparr = np.array(elm.nodes[1].coords)
        x_list.extend((p_i[0], p_j[0], None))
        y_list.extend((p_i[1], p_j[1], None))
        z_list.extend((p_i[2], p_j[2], None))
        customdata_list.append(
            (elm.uid, elm.nodes[0].uid, elm.parent_component.uid)
        )
        customdata_list.append(
            (elm.uid, elm.nodes[1].uid, elm.parent_component.uid)
        )
        customdata_list.append((None,) * 3)

    customdata: nparr = np.array(customdata_list, dtype="object")
    data_dict.append(
        {
            "name": "TwoNodeLink elements",
            "type": "scatter3d",
            "mode": "lines",
            "x": x_list,
            "y": y_list,
            "z": z_list,
            # "text": section_names,
            "customdata": customdata,
            "hovertemplate": (
                "Section: %{text}<br>"
                + "<extra>Element: %{customdata[0]:d}<br>"
                + "Node @ this end: %{customdata[1]:d}<br>"
                "Parent: %{customdata[2]}</extra>"
            ),
            "line": {"width": 5, "color": graphics_common.LINK_COLOR},
        }
    )


def add_data__frame_offsets(
    data_dict: list[dict[str, object]], mdl: Model
) -> None:
    """
    Adds a trace containing frame element rigid offset lines

    Arguments:
      data_dict: dictionary containing figure data
      mdl: the model to be visualized

    """

    beamcolumn_elems: list[element.Element] = []
    beamcolumn_elems.extend(
        mdl.list_of_specific_element(element.ElasticBeamColumn)
    )
    beamcolumn_elems.extend(
        mdl.list_of_specific_element(element.DispBeamColumn)
    )

    if not beamcolumn_elems:
        return

    x_list: list[Optional[float]] = []
    y_list: list[Optional[float]] = []
    z_list: list[Optional[float]] = []

    for elm in beamcolumn_elems:
        assert isinstance(
            elm, (element.ElasticBeamColumn, element.DispBeamColumn)
        )
        p_i: nparr = np.array(elm.nodes[0].coords)
        p_io: nparr = np.array(elm.nodes[0].coords) + elm.geomtransf.offset_i
        p_j: nparr = np.array(elm.nodes[1].coords)
        p_jo: nparr = np.array(elm.nodes[1].coords) + elm.geomtransf.offset_j

        x_list.extend((p_i[0], p_io[0], None))
        y_list.extend((p_i[1], p_io[1], None))
        z_list.extend((p_i[2], p_io[2], None))
        x_list.extend((p_j[0], p_jo[0], None))
        y_list.extend((p_j[1], p_jo[1], None))
        z_list.extend((p_j[2], p_jo[2], None))

    data_dict.append(
        {
            "name": "Rigid offsets",
            "type": "scatter3d",
            "mode": "lines",
            "x": x_list,
            "y": y_list,
            "z": z_list,
            "hoverinfo": "skip",
            "line": {"width": 8, "color": graphics_common.OFFSET_COLOR},
        }
    )


def add_data__frame_axes(
    data_dict: list[dict[str, object]], mdl: Model, ref_len: float
) -> None:
    """
    Adds a trace containing frame element local axis lines

    Arguments:
      data_dict: dictionary containing figure data
      mdl: the model to be visualized
      ref_len: model reference length to scale the axes

    """

    beamcolumn_elems: list[element.Element] = []
    beamcolumn_elems.extend(mdl.list_of_specific_element(element.TrussBar))
    beamcolumn_elems.extend(
        mdl.list_of_specific_element(element.ElasticBeamColumn)
    )
    beamcolumn_elems.extend(
        mdl.list_of_specific_element(element.DispBeamColumn)
    )
    if not beamcolumn_elems:
        return
    scaling = ref_len * 0.025
    x_list: list[Optional[float]] = []
    y_list: list[Optional[float]] = []
    z_list: list[Optional[float]] = []
    colors: list[Optional[str]] = []
    for elm in beamcolumn_elems:
        assert isinstance(
            elm, (element.ElasticBeamColumn, element.DispBeamColumn)
        )
        if elm.visibility.hidden_at_line_plots:
            continue
        x_vec = elm.geomtransf.x_axis
        y_vec = elm.geomtransf.y_axis
        z_vec = elm.geomtransf.z_axis
        l_clear = elm.clear_length()
        i_pos = np.array(elm.nodes[0].coords) + elm.geomtransf.offset_i
        mid_pos = i_pos + x_vec * l_clear / 2.00
        x_list.extend((mid_pos[0], mid_pos[0] + x_vec[0] * scaling, None))
        y_list.extend((mid_pos[1], mid_pos[1] + x_vec[1] * scaling, None))
        z_list.extend((mid_pos[2], mid_pos[2] + x_vec[2] * scaling, None))
        colors.extend(["red"] * 3)
        x_list.extend((mid_pos[0], mid_pos[0] + y_vec[0] * scaling, None))
        y_list.extend((mid_pos[1], mid_pos[1] + y_vec[1] * scaling, None))
        z_list.extend((mid_pos[2], mid_pos[2] + y_vec[2] * scaling, None))
        colors.extend(["green"] * 3)
        x_list.extend((mid_pos[0], mid_pos[0] + z_vec[0] * scaling, None))
        y_list.extend((mid_pos[1], mid_pos[1] + z_vec[1] * scaling, None))
        z_list.extend((mid_pos[2], mid_pos[2] + z_vec[2] * scaling, None))
        colors.extend(["blue"] * 3)
    data_dict.append(
        {
            "name": "Frame axes",
            "type": "scatter3d",
            "mode": "lines",
            "x": x_list,
            "y": y_list,
            "z": z_list,
            "hoverinfo": "skip",
            "line": {"width": 8, "color": colors},
        }
    )


def add_data__zerolength_axes(
    data_dict: list[dict[str, object]], mdl: Model, ref_len: float
) -> None:
    """
    Adds a trace containing zerolength element local axis lines

    Arguments:
      data_dict: dictionary containing figure data
      mdl: the model to be visualized
      ref_len: model reference length to scale the axes

    """

    zerolength_elements: list[element.Element] = []
    zerolength_elements.extend(
        mdl.list_of_specific_element(element.ZeroLength)
    )
    if not zerolength_elements:
        return
    scaling = ref_len * 0.025
    x_list: list[Optional[float]] = []
    y_list: list[Optional[float]] = []
    z_list: list[Optional[float]] = []
    colors: list[Optional[str]] = []
    for elm in zerolength_elements:
        assert isinstance(elm, element.ZeroLength)
        x_vec: nparr = elm.vecx
        y_vec: nparr = elm.vecyp
        z_vec: nparr = np.cross(x_vec, y_vec)
        mid_pos = np.array(elm.nodes[0].coords)
        x_list.extend((mid_pos[0], mid_pos[0] + x_vec[0] * scaling, None))
        y_list.extend((mid_pos[1], mid_pos[1] + x_vec[1] * scaling, None))
        z_list.extend((mid_pos[2], mid_pos[2] + x_vec[2] * scaling, None))
        colors.extend(["red"] * 3)
        x_list.extend((mid_pos[0], mid_pos[0] + y_vec[0] * scaling, None))
        y_list.extend((mid_pos[1], mid_pos[1] + y_vec[1] * scaling, None))
        z_list.extend((mid_pos[2], mid_pos[2] + y_vec[2] * scaling, None))
        colors.extend(["green"] * 3)
        x_list.extend((mid_pos[0], mid_pos[0] + z_vec[0] * scaling, None))
        y_list.extend((mid_pos[1], mid_pos[1] + z_vec[1] * scaling, None))
        z_list.extend((mid_pos[2], mid_pos[2] + z_vec[2] * scaling, None))
        colors.extend(["blue"] * 3)
    data_dict.append(
        {
            "name": "Zerolength axes",
            "type": "scatter3d",
            "mode": "lines",
            "x": x_list,
            "y": y_list,
            "z": z_list,
            "hoverinfo": "skip",
            "line": {"width": 8, "color": colors},
        }
    )


def add_data__global_axes(
    data_dict: list[dict[str, object]], ref_len: float
) -> None:
    """
    Adds a trace containing global axes

    Arguments:
      data_dict: dictionary containing figure data
      ref_len: model reference length to scale the axes

    """

    scaling = ref_len
    x_list: list[Optional[float]] = []
    y_list: list[Optional[float]] = []
    z_list: list[Optional[float]] = []
    colors: list[Optional[str]] = []
    x_vec: nparr = np.array([1.00, 0.00, 0.00])
    y_vec: nparr = np.array([0.00, 1.00, 0.00])
    z_vec: nparr = np.array([0.00, 0.00, 1.00])

    x_list.extend((0.00, x_vec[0] * scaling, None))
    y_list.extend((0.00, x_vec[1] * scaling, None))
    z_list.extend((0.00, x_vec[2] * scaling, None))
    x_list.extend((0.00, y_vec[0] * scaling, None))
    y_list.extend((0.00, y_vec[1] * scaling, None))
    z_list.extend((0.00, y_vec[2] * scaling, None))
    x_list.extend((0.00, z_vec[0] * scaling, None))
    y_list.extend((0.00, z_vec[1] * scaling, None))
    z_list.extend((0.00, z_vec[2] * scaling, None))
    colors.extend(["red"] * 3)
    colors.extend(["green"] * 3)
    colors.extend(["blue"] * 3)
    # we add it twice in order for the animations to work
    # see https://plotly.com/python/animations/
    data_dict.extend(
        [
            {
                "type": "scatter3d",
                "mode": "lines",
                "x": x_list,
                "y": y_list,
                "z": z_list,
                "hoverinfo": "skip",
                "line": {"width": 3, "color": colors},
            }
        ]
        * 2
    )


def add_data__diaphragm_lines(
    data_dict: list[dict[str, object]], loadcase: LoadCase
) -> None:
    """
    Adds a trace containing lines indicating rigid diaphragm
    constraints.

    Arguments:
      data_dict: dictionary containing figure data
      loadcase: loadcase

    """

    # if parent nodes don't exist, we don't have rigid diaphragms
    pnodes = loadcase.parent_nodes
    if not pnodes:
        return

    # otherwise, rigid diaphragms exist and we will plot lines, so we
    # prepare containers
    x_list: list[Optional[float]] = []
    y_list: list[Optional[float]] = []
    z_list: list[Optional[float]] = []

    # for each level that has a parent node, draw the lines
    mdl = loadcase.parent_model
    for lvl_uid, pnode in pnodes.items():
        level = mdl.levels[lvl_uid]
        level_primary_nodes = level.nodes.values()
        for node in level_primary_nodes:
            if node.coords[2] == level.elevation:
                x_list.extend((node.coords[0], pnode.coords[0], None))
                y_list.extend((node.coords[1], pnode.coords[1], None))
                z_list.extend((level.elevation, level.elevation, None))

    data_dict.append(
        {
            "name": "Rigid Diaphragms",
            "type": "scatter3d",
            "mode": "lines",
            "x": x_list,
            "y": y_list,
            "z": z_list,
            "hoverinfo": "skip",
            "line": {"width": 2, "color": graphics_common.GRID_COLOR},
        }
    )


def add_data__tributary_area_boundaries(
    data_dict: list[dict[str, object]], loadcase: LoadCase
) -> None:
    """
    Adds a trace containing lines indicating tributary area boundaries
    that are used for load distribution.

    Arguments:
      data_dict: dictionary containing figure data
      loadcase: loadcase

    """

    trib_area_analysis_collection = loadcase.tributary_area_analysis

    x_list: list[Optional[float]] = []
    y_list: list[Optional[float]] = []
    z_list: list[Optional[float]] = []

    for analysis in trib_area_analysis_collection.values():
        level = analysis.parent_level
        data = analysis.data
        edge_polygons = data.edge_polygons

        for polygon_list in edge_polygons.values():
            for polygon in polygon_list:
                x_vals = [v[0] for v in polygon]
                y_vals = [v[1] for v in polygon]
                x_vals.append(x_vals[0])
                y_vals.append(y_vals[0])
                x_list.extend((*x_vals, None))
                y_list.extend((*y_vals, None))
                z_list.extend((*[level.elevation] * len(x_vals), None))

    data_dict.append(
        {
            "name": "Trib. Area",
            "type": "scatter3d",
            "mode": "lines",
            "x": x_list,
            "y": y_list,
            "z": z_list,
            "hoverinfo": "skip",
            "line": {"width": 2, "color": graphics_common.BISECTOR_COLOR},
        }
    )


def add_data__extruded_frames_mesh(
    data_dict: list[dict[str, object]], mdl: Model
) -> None:
    """
    Adds a trace containing frame element extrusion mesh

    Arguments:
      data_dict: dictionary containing figure data
      mdl: the model to be visualized

    """

    beamcolumn_elems: list[element.Element] = []
    beamcolumn_elems.extend(
        mdl.list_of_specific_element(element.ElasticBeamColumn)
    )
    beamcolumn_elems.extend(
        mdl.list_of_specific_element(element.DispBeamColumn)
    )
    if not beamcolumn_elems:
        return
    x_list: list[Optional[float]] = []
    y_list: list[Optional[float]] = []
    z_list: list[Optional[float]] = []
    i_list: list[Optional[int]] = []
    j_list: list[Optional[int]] = []
    k_list: list[Optional[int]] = []
    index = 0
    for elm in beamcolumn_elems:
        assert isinstance(
            elm, (element.ElasticBeamColumn, element.DispBeamColumn)
        )
        if elm.visibility.hidden_when_extruded:
            continue
        side_a = np.array(elm.nodes[0].coords) + elm.geomtransf.offset_i
        side_b = np.array(elm.nodes[1].coords) + elm.geomtransf.offset_j
        y_vec = elm.geomtransf.y_axis
        z_vec = elm.geomtransf.z_axis
        if not elm.section.outside_shape:
            continue
        loop = elm.section.outside_shape.halfedges
        for halfedge in loop:
            assert halfedge.nxt is not None
            assert halfedge.nxt.vertex is not None
            loc0 = (
                halfedge.vertex.coords[0] * z_vec
                + halfedge.vertex.coords[1] * y_vec
                + side_a
            )
            loc1 = (
                halfedge.vertex.coords[0] * z_vec
                + halfedge.vertex.coords[1] * y_vec
                + side_b
            )
            loc2 = (
                halfedge.nxt.vertex.coords[0] * z_vec
                + halfedge.nxt.vertex.coords[1] * y_vec
                + side_b
            )
            loc3 = (
                halfedge.nxt.vertex.coords[0] * z_vec
                + halfedge.nxt.vertex.coords[1] * y_vec
                + side_a
            )
            x_list.extend((loc0[0], loc1[0], loc2[0], loc3[0]))
            y_list.extend((loc0[1], loc1[1], loc2[1], loc3[1]))
            z_list.extend((loc0[2], loc1[2], loc2[2], loc3[2]))
            i_list.extend((index + 0, index + 0))
            j_list.extend((index + 1, index + 2))
            k_list.extend((index + 2, index + 3))
            index += 4
    data_dict.append(
        {
            "type": "mesh3d",
            "x": x_list,
            "y": y_list,
            "z": z_list,
            "i": i_list,
            "j": j_list,
            "k": k_list,
            "hoverinfo": "skip",
            "color": graphics_common.BEAM_MESH_COLOR,
            "opacity": 0.70,
        }
    )


def add_data__extruded_bars_mesh(
    data_dict: list[dict[str, object]], mdl: Model
) -> None:
    """
    Adds a trace containing frame element extrusion mesh

    Arguments:
      data_dict: dictionary containing figure data
      mdl: the model to be visualized

    """

    line_elems = mdl.list_of_specific_element(element.TrussBar)
    if not line_elems:
        return
    x_list: list[Optional[float]] = []
    y_list: list[Optional[float]] = []
    z_list: list[Optional[float]] = []
    i_list: list[Optional[int]] = []
    j_list: list[Optional[int]] = []
    k_list: list[Optional[int]] = []
    index = 0
    for elm in line_elems:
        assert isinstance(elm, element.TrussBar)
        if elm.visibility.hidden_when_extruded:
            continue
        if not elm.outside_shape:
            continue

        side_a = np.array(elm.nodes[0].coords)
        side_b = np.array(elm.nodes[1].coords)
        x_vec, y_vec, z_vec = local_axes_from_points_and_angle(
            side_a, side_b, 0.00
        )

        # cut out the two ends for a nicer visual effect
        # (after all we can't plot the exact connection geometry)
        elm_len = elm.clear_length()
        side_a += x_vec * elm_len * 0.33
        side_b -= x_vec * elm_len * 0.33

        loop = elm.outside_shape.halfedges
        for halfedge in loop:
            assert halfedge.nxt
            loc0 = (
                halfedge.vertex.coords[0] * z_vec
                + halfedge.vertex.coords[1] * y_vec
                + side_a
            )
            loc1 = (
                halfedge.vertex.coords[0] * z_vec
                + halfedge.vertex.coords[1] * y_vec
                + side_b
            )
            loc2 = (
                halfedge.nxt.vertex.coords[0] * z_vec
                + halfedge.nxt.vertex.coords[1] * y_vec
                + side_b
            )
            loc3 = (
                halfedge.nxt.vertex.coords[0] * z_vec
                + halfedge.nxt.vertex.coords[1] * y_vec
                + side_a
            )
            x_list.extend((loc0[0], loc1[0], loc2[0], loc3[0]))
            y_list.extend((loc0[1], loc1[1], loc2[1], loc3[1]))
            z_list.extend((loc0[2], loc1[2], loc2[2], loc3[2]))
            i_list.extend((index + 0, index + 0))
            j_list.extend((index + 1, index + 2))
            k_list.extend((index + 2, index + 3))
            index += 4
    data_dict.append(
        {
            "type": "mesh3d",
            "x": x_list,
            "y": y_list,
            "z": z_list,
            "i": i_list,
            "j": j_list,
            "k": k_list,
            "hoverinfo": "skip",
            "color": graphics_common.BEAM_MESH_COLOR,
            "opacity": 0.30,
        }
    )


def show(
    mdl: Model,
    load_case: Optional[LoadCase] = None,
    extrude: bool = False,
    offsets: bool = True,
    global_axes: bool = True,
    parent_nodes: bool = True,
    frame_axes: bool = False,
    zerolength_axes: bool = False,
    diaphragm_lines: bool = True,
    tributary_area_boundaries: bool = True,
    camera: Optional[dict[str, object]] = None,
    to_html_file: Optional[str] = None,
    to_figure: Optional[str] = None,
) -> None:
    """
    Visualize the model

    Arguments:
      mdl: the model to be visualized
      load_case: the load_case to be visualized
      extrude: wether to extrude frame elements
      offsets: whether to show frame element rigid offsets
      global_axes: whether to show global axes
      diaphragm_lines: whether to show lines indicating rigid
        diaphragm extent
      tributary_area_boundaries: whether to show tributary area
        boundary lines
      parent_nodes: whether to plot parent nodes
      frame_axes: whether to show the local axes of frame
        elements
      zerolength_axes: whether to show the local axes of
        zerolength elements
      camera: custom positioning of the camera
      to_html_file: If a path is specified, the figure is written in
        an html file instead of being shown.
      to_figure: If a path is specified, the figure is
        saved in the specified format.
    """

    layout = graphics_common_3d.global_layout(mdl, camera)
    data_dict: list[dict[str, object]] = []

    ref_len = mdl.reference_length()

    # plot the nodes

    add_data__nodes(data_dict, mdl, load_case)
    add_data__internal_nodes(data_dict, mdl, load_case)

    # global axes
    if global_axes:
        add_data__global_axes(data_dict, ref_len)

    # lines indicating rigid diaphragms
    if load_case:
        if diaphragm_lines:
            add_data__diaphragm_lines(data_dict, load_case)

    # tributary area boundary lines
    if load_case:
        if tributary_area_boundaries:
            add_data__tributary_area_boundaries(data_dict, load_case)

    # plot beamcolumn elements
    add_data__frames(data_dict, mdl, load_case)
    add_data__bars(data_dict, mdl, load_case)
    if frame_axes:
        add_data__frame_axes(data_dict, mdl, ref_len)
    if zerolength_axes:
        add_data__zerolength_axes(data_dict, mdl, ref_len)
    if extrude:
        add_data__extruded_frames_mesh(data_dict, mdl)
        add_data__extruded_bars_mesh(data_dict, mdl)
    # plot the rigid offsets
    if offsets:
        add_data__frame_offsets(data_dict, mdl)
    add_data__twonodelinks(data_dict, mdl)
    # plot the parent nodes
    if parent_nodes:
        if load_case:
            add_data__parent_nodes(data_dict, load_case)

    fig_datastructure = dict(data=data_dict, layout=layout)
    fig = go.Figure(fig_datastructure)

    if "pytest" not in sys.modules:
        if to_html_file:
            fig.write_html(to_html_file)
        elif to_figure:
            pio.write_image(fig, to_figure, width=2000, height=2000)
        else:
            fig.show()



"""
Objects that generate materials.

"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from dataclasses import dataclass
import numpy as np
import numpy.typing as npt
from ..ops.section import FiberSection
from ..physical_material import PhysicalMaterial
from ..model import Model
from ..ops.uniaxial_material import Elastic
from ..ops.uniaxial_material import Steel02
from ..ops.uniaxial_material import Fatigue
from ..ops.uniaxial_material import MaxStrainRange
from ..ops.uniaxial_material import IMKBilin
from ..ops.section import ElasticSection

nparr = npt.NDArray[np.float64]


@dataclass(repr=False)
class MaterialGenerator:
    """
    This object introduces element-specific materials to a model.
    """

    model: Model

    def generate_steel_hss_rect_brace_maxstrainrange_mat(
        self,
        section: FiberSection,
        physical_material: PhysicalMaterial,
        brace_length: float,
        node_i_uid: int,
        node_j_uid: int,
    ) -> MaxStrainRange:
        """
        Sen, A. D., Roeder, C. W., Lehman, D. E., & Berman,
        J. W. (2019). Nonlinear modeling of concentrically braced
        frames. Journal of Constructional Steel Research, 157,
        103-120.

        """

        param_b = 0.001
        param_r0 = 15
        param_c_r1 = 0.925
        param_c_r2 = 0.15

        assert section.properties is not None
        sec_b = section.properties["B"]
        sec_t = section.properties["tdes"]
        var_lc = brace_length
        sec_r = min(section.properties["rx"], section.properties["ry"])
        mat_e = physical_material.e_mod
        mat_g = physical_material.g_mod
        mat_fy = physical_material.f_y
        var_msr = (
            (0.554)
            * (sec_b / sec_t) ** (-0.75)
            * (var_lc / sec_r) ** (-0.47)
            * (mat_e / mat_fy) ** (0.21)
        )

        steel02_mat = Steel02(
            self.model.uid_generator.new("uniaxial material"),
            "auto_steel02_brace_mat",
            mat_fy,
            mat_e,
            mat_g,
            param_b,
            param_r0,
            param_c_r1,
            param_c_r2,
        )

        maxstrainrange_mat = MaxStrainRange(
            self.model.uid_generator.new("uniaxial material"),
            "auto_maxstrainrange_brace_mat",
            steel02_mat,
            var_msr,
            tangent_ratio=1.0e-5,
            def_coeff=0.068,
            node_tags=(node_i_uid, node_j_uid),
        )

        return maxstrainrange_mat

    def generate_steel_hss_circ_brace_fatigue_mat(
        self,
        section: FiberSection,
        physical_material: PhysicalMaterial,
        brace_length: float,
    ) -> Fatigue:
        """
        Karamanaci and Lignos (2014). Computational Approach for
        Collapse Assessment of Concentrically Braced Frames in Seismic
        Regions. Journal of Structural Engineering

        """

        param_b = 0.005
        param_r0 = 24.00
        param_c_r1 = 0.925
        param_c_r2 = 0.25

        assert section.properties is not None
        sec_d = section.properties["OD"]
        sec_t = section.properties["tdes"]
        var_lc = brace_length
        sec_r = min(section.properties["rx"], section.properties["ry"])
        mat_e = physical_material.e_mod
        mat_g = physical_material.g_mod
        mat_fy = physical_material.f_y
        var_e0 = (
            (0.748)
            * (var_lc / sec_r) ** (-0.399)
            * (sec_d / sec_t) ** (-0.628)
            * (mat_e / mat_fy) ** (0.201)
        )
        var_m = -0.300

        steel02_mat = Steel02(
            self.model.uid_generator.new("uniaxial material"),
            "auto_steel02_brace_mat",
            mat_fy,
            mat_e,
            mat_g,
            param_b,
            param_r0,
            param_c_r1,
            param_c_r2,
            a1=0.2,
            a2=1.0,
            a3=0.2,
            a4=1.0,
        )

        fatigue_mat = Fatigue(
            self.model.uid_generator.new("uniaxial material"),
            "auto_fatigue_brace_mat",
            steel02_mat,
            var_e0,
            var_m,
        )

        return fatigue_mat

    def generate_steel_w_imk_material(
        self,
        section,
        physical_material,
        element_length,
        lboverl,
        loverh,
        rbs_factor,
        consider_composite,
        axial_load_ratio,
        direction="strong",
        moment_modifier=1.00,
        n_parameter=0.00,
        only_elastic=False,
    ):
        """
        Lignos, D. G., & Krawinkler, H. (2011). Deterioration modeling of
        steel components in support of collapse prediction of steel moment
        frames under earthquake loading. Journal of Structural
        Engineering-Reston, 137(11), 1291.

        Elkady, A., & Lignos, D. G. (2014). Modeling of the composite
        action in fully restrained beam‐to‐column connections:
        implications in the seismic design and collapse capacity of steel
        special moment frames. Earthquake Engineering & Structural
        Dynamics, 43(13), 1935-1954.

        """

        # gather necessary data and check interpolation range
        assert section.name[0] == "W", "Error: Only W sections can be used."
        assert isinstance(section, ElasticSection)
        assert (
            self.model.settings.imperial_units
        ), "Error: Only imperial units supported."
        assert section.properties
        # Yield stress
        mat_fy = physical_material.f_y / 1.0e3
        # Moment of inertia - strong axis - original section
        if direction == "strong":
            sec_i = section.properties["Ix"]
        else:
            sec_i = section.properties["Iy"]
        # Section depth
        sec_d = section.properties["d"]
        # Flange width
        sec_bf = section.properties["bf"]
        # Flange and web thicknesses
        sec_tf = section.properties["tf"]
        sec_tw = section.properties["tw"]
        # Plastic modulus (unreduced)
        if direction == "strong":
            sec_z = section.properties["Zx"]
        else:
            sec_z = section.properties["Zy"]
        # Radius of gyration
        sec_ry = section.properties["ry"]
        # Clear length
        elm_h = element_length
        # Shear span
        elm_l = loverh * elm_h
        elm_lb = lboverl * elm_l
        lbry = elm_lb / sec_ry

        # consider cases

        if rbs_factor:
            # RBS case
            assert direction == "strong"
            assert rbs_factor <= 1.00, "rbs_factor must be <= 1.00"
            # checks ~ acceptable range
            if not 20.00 < sec_d / sec_tw < 55.00:
                print(
                    f"Warning: sec_d/sec_tw={sec_d/sec_tw:.2f}"
                    " outside regression range"
                )
                print("20.00 < sec_d/sec_tw < 55.00")
                print(section.name, "\n")
            if not 20.00 < lbry < 80.00:
                print(f"Warning: Lb/ry={lbry:.2f} outside regression range")
                print("20.00 < lbry < 80.00")
                print(section.name, "\n")
            if not 4.00 < (sec_bf / (2.0 * sec_tf)) < 8.00:
                print(
                    f"Warning: bf/(2 tf)={sec_bf/(2.*sec_tf):.2f}"
                    " outside regression range"
                )
                print("4.00 < (sec_bf/(2.*sec_tf)) < 8.00")
                print(section.name, "\n")
            if not 2.5 < elm_l / sec_d < 7.0:
                print(
                    f"Warning: L/d={elm_l/sec_d:.2f}  outside regression range"
                )
                print("2.5 < elm_l/sec_d < 7.0")
                print(section.name, "\n")
            if not 4.00 < sec_d < 36.00:
                print(
                    f"Warning: Section d={sec_d:.2f} outside regression range"
                )
                print("4.00 < sec_d < 36.00")
                print(section.name, "\n")
            if not 35.00 < mat_fy < 65.00:
                print(f"Warning: Fy={mat_fy:.2f} outside regression range")
                print("35.00 < mat_fy < 65.00")
                print(section.name, "\n")
            # calculate parameters
            theta_p = (
                0.19
                * (sec_d / sec_tw) ** (-0.314)
                * (sec_bf / (2.0 * sec_tf)) ** (-0.10)
                * lbry ** (-0.185)
                * (elm_l / sec_d) ** 0.113
                * (25.4 * sec_d / 533.0) ** (-0.76)
                * (6.895 * mat_fy / 355.0) ** (-0.07)
            )
            theta_pc = (
                9.52
                * (sec_d / sec_tw) ** (-0.513)
                * (sec_bf / (2.0 * sec_tf)) ** (-0.863)
                * lbry ** (-0.108)
                * (6.895 * mat_fy / 355.0) ** (-0.36)
            )
            lamda = (
                585.0
                * (sec_d / sec_tw) ** (-1.14)
                * (sec_bf / (2.0 * sec_tf)) ** (-0.632)
                * lbry ** (-0.205)
                * (6.895 * mat_fy / 355.0) ** (-0.391)
            )
            rbs_c = sec_bf * (1.0 - rbs_factor) / 2.0
            z_rbs = sec_z - 2.0 * rbs_c * sec_tf * (sec_d - sec_tf)
            sec_m = 1.06 * z_rbs * mat_fy * 1.0e3
            mcmy_plus = 1.10
            mcmy_minus = 1.10

        else:
            # Other-than-RBS case
            if axial_load_ratio:
                # column case
                theta_p = (
                    294.00
                    * (sec_d / sec_tw) ** (-1.70)
                    * lbry ** (-0.70)
                    * (1.00 - axial_load_ratio) ** (1.60)
                )
                theta_pc = (
                    90.00
                    * (sec_d / sec_tw) ** (-0.80)
                    * lbry ** (-0.80)
                    * (1.00 - axial_load_ratio) ** (2.50)
                )
                theta_p = min(theta_p, 0.20)
                theta_pc = min(theta_pc, 0.30)
                if axial_load_ratio <= 0.35:
                    lamda = (
                        25500.00
                        * (sec_d / sec_tw) ** (-2.14)
                        * lbry ** (-0.53)
                        * (1.00 - axial_load_ratio) ** (4.29)
                    )
                else:
                    lamda = (
                        268000.00
                        * (sec_d / sec_tw) ** (-2.30)
                        * lbry ** (-1.30)
                        * (1.00 - axial_load_ratio) ** (1.19)
                    )
                if axial_load_ratio <= 0.20:
                    sec_m = (
                        1.15
                        / 1.10
                        * (sec_z * mat_fy)
                        * 1.0e3
                        * (1.00 - axial_load_ratio / 2.00)
                    )
                else:
                    sec_m = (
                        1.15
                        / 1.10
                        * (sec_z * mat_fy)
                        * 1.0e3
                        * 9.0
                        / 8.0
                        * (1.00 - axial_load_ratio)
                    )
                mcmy = (
                    12.5
                    * (sec_d / sec_tw) ** (-0.20)
                    * lbry ** (-0.40)
                    * (1.00 - axial_load_ratio) ** (0.40)
                )
                mcmy = min(mcmy, 1.00)
                mcmy = max(mcmy, 1.30)
                mcmy_plus = mcmy
                mcmy_minus = mcmy

            else:
                # non-RBS beam case
                theta_p = (
                    0.0865
                    * (sec_d / sec_tw) ** (-0.365)
                    * (sec_bf / (2.0 * sec_tf)) ** (-0.14)
                    * (elm_l / sec_d) ** 0.34
                    * (25.4 * sec_d / 533.0) ** (-0.721)
                    * (6.895 * mat_fy / 355.0) ** (-0.23)
                )
                theta_pc = (
                    5.63
                    * (sec_d / sec_tw) ** (-0.565)
                    * (sec_bf / (2.0 * sec_tf)) ** (-0.800)
                    * (25.4 * sec_d / 533.0) ** (-0.28)
                    * (6.895 * mat_fy / 355.0) ** (-0.43)
                )
                lamda = (
                    495.0
                    * (sec_d / sec_tw) ** (-1.34)
                    * (sec_bf / (2.0 * sec_tf)) ** (-0.595)
                    * (6.895 * mat_fy / 355.0) ** (-0.36)
                )
                sec_m = 1.17 * sec_z * mat_fy * 1.0e3
                mcmy_plus = 1.10
                mcmy_minus = 1.10

        theta_u = 0.20
        residual_plus = 0.40
        residual_minus = 0.40
        theta_p_plus = theta_p
        theta_p_minus = theta_p
        theta_pc_plus = theta_pc
        theta_pc_minus = theta_pc
        d_plus = 1.00
        d_minus = 1.00
        m_plus = sec_m
        m_minus = -sec_m

        if consider_composite:
            # Elkady, A., & Lignos, D. G. (2014). Modeling of the
            # composite action in fully restrained beam‐to‐column
            # connections: implications in the seismic design and
            # collapse capacity of steel special moment
            # frames. Earthquake Engineering & Structural Dynamics,
            # 43(13), 1935-1954.  Table II

            assert (
                axial_load_ratio == 0.00
            ), "Can't consider composite action for columns"
            assert (
                direction == "strong"
            ), "Composite action affects the behavior in strong-axis bending"

            theta_p_plus *= 1.80
            theta_p_minus *= 0.95
            theta_pc_plus *= 1.35
            theta_pc_minus *= 0.95
            d_plus = 1.15
            d_minus = 1.00
            mcmy_plus = 1.30
            mcmy_minus = 1.05
            m_plus *= 1.35
            m_minus *= 1.25
            residual_plus = 0.30
            residual_minus = 0.20

        # adjust parameters to account for the presence of the elastic element
        stiffness_init = 6.00 * section.e_mod * sec_i / elm_h
        stiffness = (n_parameter + 1.00) * stiffness_init
        theta_y = sec_m / stiffness_init
        theta_p_plus -= (mcmy_plus - 1.0) * (sec_m / stiffness)
        theta_p_minus -= (mcmy_minus - 1.0) * (sec_m / stiffness)
        theta_pc_plus += theta_y + (mcmy_plus - 1.0) * (sec_m / stiffness)
        theta_pc_plus += theta_y + (mcmy_minus - 1.0) * (sec_m / stiffness)
        beta_plus = (mcmy_plus - 1.0) * m_plus / theta_p_plus / stiffness
        beta_minus = -(mcmy_minus - 1.0) * m_minus / theta_p_minus / stiffness

        # # old model
        # from ..ops.uniaxial_material import Bilin
        # bilin_mat = Bilin(
        #     self.model.uid_generator.new("uniaxial material"),
        #     "auto_IMK",
        #     stiffness * moment_modifier,
        #     beta_plus,
        #     beta_minus,
        #     m_plus*moment_modifier,
        #     m_minus*moment_modifier,
        #     lamda,
        #     lamda,
        #     lamda,
        #     lamda,
        #     1.00,
        #     1.00,
        #     1.00,
        #     1.00,
        #     theta_p_plus,
        #     theta_p_minus,
        #     theta_pc_plus,
        #     theta_pc_minus,
        #     residual_plus,
        #     residual_minus,
        #     theta_u,
        #     theta_u,
        #     d_plus,
        #     d_minus,
        #     0.00,
        # )

        # new model
        if only_elastic:
            elastic_mat = Elastic(
                self.model.uid_generator.new("uniaxial material"),
                "auto_IMK",
                stiffness * moment_modifier,
            )
            return elastic_mat
        bilin_mat = IMKBilin(
            self.model.uid_generator.new("uniaxial material"),
            "auto_IMK",
            stiffness * moment_modifier,
            theta_p_plus,
            theta_pc_plus,
            theta_u,
            m_plus * moment_modifier,
            (1.0 + beta_plus),
            residual_plus,
            theta_p_minus,
            theta_pc_minus,
            theta_u,
            -m_minus * moment_modifier,
            (1.0 + beta_minus),
            residual_minus,
            lamda,
            lamda,
            lamda,
            1.00,
            1.00,
            1.00,
            d_plus,
            d_minus,
        )
        return bilin_mat



"""
Defines objects associated with tributary area analysis for load
distribution.

"""


#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from typing import TYPE_CHECKING
from typing import no_type_check
from typing import Optional
from dataclasses import dataclass, field
import sys
from tqdm import tqdm
import numpy as np
import numpy.typing as npt
from .. import mesh
from ..ops.node import Node
from .. import common
from ..ops.element import ElasticBeamColumn
from ..ops.element import DispBeamColumn
from ..ops.element import TwoNodeLink
from ..ops.element import ZeroLength

if TYPE_CHECKING:
    from ..load_case import LoadCase
    from ..level import Level
    from ..mesh import Edge, Vertex, Halfedge

# pylint: disable=no-member
# pylint: disable=import-outside-toplevel

nparr = npt.NDArray[np.float64]
PolygonShape = list[tuple[float, float]]


@dataclass
class PolygonLoad:
    """
    Polygon-shaped loaded region.

    """

    name: str
    value: float
    outside_shape: Optional[PolygonShape]
    holes: Optional[list[PolygonShape]]
    massless: bool


@dataclass(repr=False)
class TributaryAreaAnalysisData:
    """
    Data generated by the tributary area analysis code.

    """

    # collects the defined edges
    edges: dict[int, Edge] = field(default_factory=dict)
    # collects the defined vertices
    vertices: dict[int, Vertex] = field(default_factory=dict)
    # maps an edge.uid to a node or line element
    edge_map: dict[int, Node | ElasticBeamColumn | DispBeamColumn] = field(
        default_factory=dict
    )
    # maps an edge.uid to its corresponding tributary area
    edge_area: dict[int, float] = field(default_factory=dict)
    # maps an edge.uid to its corresponding tributary polygon
    edge_polygons: dict[int, list[PolygonShape]] = field(default_factory=dict)
    # maps a node.uid to a vertex
    vertex_map: dict[int, Vertex] = field(default_factory=dict)
    # used to map zerolength element-induced coinciding nodes
    zn_map: dict[int, int] = field(default_factory=dict)
    # used to map panel zone nodes to their primary node
    pz_node: dict[int, dict[str, Node]] = field(default_factory=dict)


@dataclass
class TributaryAreaAnaysis:
    """
    Tributary area analysis object.

    """

    parent_loadcase: LoadCase
    parent_level: Level
    polygon_loads: list[PolygonLoad] = field(default_factory=list)
    data: TributaryAreaAnalysisData = field(
        default_factory=TributaryAreaAnalysisData
    )

    @no_type_check
    def run(
        self, load_factor=1.00, massless_load_factor=1.00, perform_checks=True
    ):
        """
        Performs tributary area analysis

        Arguments:
          load_factor: Factor by which to multiply the surface
            loads that count as a mass source, found in the load case.
          massless_load_factor: Factor by which to multiply
            the massless loads found in the load case.
          perform_checks: Tributary area analysis works by
            generating a mesh defined by the horizontal elements of
            the floor plan and analyzing that mesh. The mesh should
            not contain any edges that cross or overlap with other
            edges. This can happen, for instance, if beam end offsets
            have been accidentally misspecified causing beams to
            overlap with or cross other members. The only shared
            points between the edges should be the end vertices. By
            default, we check if that is the case to avoid unexpected
            behavior. This flag can be used to disable the checks for
            subsequent analyses to enhance performance if that a prior
            analysis has verified that the checks are satisfied.

        """

        try:
            import skgeom as sg
        except ModuleNotFoundError:
            msg = "One day, a custom implementation of the "
            msg += "straight skeleton algorithm might be added to osmg...\n"
            msg += "Until that day, the scikit-geometry package is required.\n"
            msg += "Please install scikit-geometry.\n"
            msg += "  $ conda install scikit-geometry -c conda-forge\n"
            msg += "https://github.com/scikit-geometry/scikit-geometry"
            print("~ skgeom is not installed ~")
            print()
            print(msg)
            sys.exit()

        lvl = self.parent_level
        all_components = list(lvl.components.values())
        horizontal_elements = []
        panel_zones = []
        for component in all_components:
            if component.component_purpose == "horizontal_component":
                horizontal_elements.append(component)
            elif component.component_purpose == "steel_W_panel_zone":
                panel_zones.append(component)

        # # plotting - used while developing the code
        # subset_model = mdl.initialize_empty_copy('subset_1')
        # for component in horizontal_elements+panel_zones:
        #     mdl.transfer_component(subset_model, component)
        # show(subset_model)

        # define vertices, edges, mapping
        # we "flatten" the components in the xy plane.
        # for joints, we only consider the top-most line elements
        # and we treat them the same as rigid offsets.

        # clear any previous results
        self.data.edges = {}
        self.data.vertices = {}
        self.data.edge_map = {}
        self.data.edge_area = {}
        self.data.edge_polygons = {}
        self.data.vertex_map = {}
        self.data.zn_map = {}
        self.data.pz_node = {}

        edges = self.data.edges
        vertices = self.data.vertices
        edge_map = self.data.edge_map
        edge_area = self.data.edge_area
        edge_polygons = self.data.edge_polygons
        vertex_map = self.data.vertex_map
        zn_map = self.data.zn_map
        pz_node = self.data.pz_node

        for panel_zone in panel_zones:
            back_nd = panel_zone.external_nodes.named_contents["middle_back"]
            front_nd = panel_zone.external_nodes.named_contents["middle_front"]
            main_nd = panel_zone.external_nodes.named_contents["top_node"]
            back_offset: nparr = np.array(main_nd.coords) - np.array(
                back_nd.coords
            )
            front_offset: nparr = np.array(main_nd.coords) - np.array(
                front_nd.coords
            )
            pz_node[back_nd.uid] = {
                "substitute_node": main_nd,
                "additional_offset": -back_offset[0:2],
            }
            pz_node[front_nd.uid] = {
                "substitute_node": main_nd,
                "additional_offset": -front_offset[0:2],
            }

        for comp in horizontal_elements:
            line_elements = [
                elm
                for elm in comp.elements.values()
                if isinstance(
                    elm, (ElasticBeamColumn, DispBeamColumn, TwoNodeLink)
                )
            ]

            zerolength_elements = [
                elm
                for elm in comp.elements.values()
                if isinstance(elm, ZeroLength)
            ]
            for zelm in zerolength_elements:
                zn_map[zelm.nodes[1].uid] = zelm.nodes[0].uid

            for elm in line_elements:
                n_i = elm.nodes[0]
                if hasattr(elm, 'geomtransf'):
                    eo_i = elm.geomtransf.offset_i[0:2]
                else:
                    eo_i = np.zeros(2)
                if n_i.uid in zn_map:
                    n_i = comp.internal_nodes[zn_map[n_i.uid]]
                n_j = elm.nodes[1]
                if hasattr(elm, 'geomtransf'):
                    eo_j = elm.geomtransf.offset_j[0:2]
                else:
                    eo_j = np.zeros(2)
                if n_j.uid in zn_map:
                    n_j = comp.internal_nodes[zn_map[n_j.uid]]
                if n_i.uid in pz_node:
                    sub_data = pz_node[n_i.uid]
                    n_i = sub_data["substitute_node"]
                    eo_i = eo_i.copy() + sub_data["additional_offset"]
                if n_j.uid in pz_node:
                    sub_data = pz_node[n_j.uid]
                    n_j = sub_data["substitute_node"]
                    eo_j = eo_j.copy() + sub_data["additional_offset"]

                if n_i.uid not in vertex_map:
                    vrt_i = mesh.Vertex((n_i.coords[0], n_i.coords[1]))
                    vertex_map[n_i.uid] = vrt_i
                    vertices[vrt_i.uid] = vrt_i
                else:
                    vrt_i = vertex_map[n_i.uid]
                if n_j.uid not in vertex_map:
                    vrt_j = mesh.Vertex((n_j.coords[0], n_j.coords[1]))
                    vertex_map[n_j.uid] = vrt_j
                    vertices[vrt_j.uid] = vrt_j
                else:
                    vrt_j = vertex_map[n_j.uid]

                if np.linalg.norm(eo_i) >= common.EPSILON:
                    # there is a rigid offset and/or panel zone
                    point = np.array(vrt_i.coords) + eo_i
                    vrt_oi = mesh.Vertex((point[0], point[1]))
                    vertices[vrt_oi.uid] = vrt_oi
                    edg_oi = mesh.Edge(vrt_i, vrt_oi)
                    edges[edg_oi.uid] = edg_oi
                    edge_map[edg_oi.uid] = n_i
                    connecting_vertex_i = vrt_oi
                else:
                    connecting_vertex_i = vrt_i
                if np.linalg.norm(eo_j) >= common.EPSILON:
                    # there is a rigid offset and/or panel zone
                    point = np.array(vrt_j.coords) + eo_j
                    vrt_oj = mesh.Vertex((point[0], point[1]))
                    vertices[vrt_oj.uid] = vrt_oj
                    edg_oj = mesh.Edge(vrt_j, vrt_oj)
                    edges[edg_oj.uid] = edg_oj
                    edge_map[edg_oj.uid] = n_j
                    connecting_vertex_j = vrt_oj
                else:
                    connecting_vertex_j = vrt_j
                edg_interior = mesh.Edge(
                    connecting_vertex_i, connecting_vertex_j
                )
                edge_map[edg_interior.uid] = elm
                edges[edg_interior.uid] = edg_interior

        # # plotting - used while developing the code
        # import pandas as pd
        # edf = pd.DataFrame(np.zeros((len(edges)*3, 2)))
        # edf.columns = ['x', 'y']
        # enames = []
        # for i, edge_key in enumerate(edges):
        #     edge = edges[edge_key]
        #     edf.loc[i*3+0, 'x':'y'] = edge.v_i.coords
        #     edf.loc[i*3+1, 'x':'y'] = edge.v_j.coords
        #     edf.loc[i*3+2, 'x':'y'] = (None, None)
        #     enames.extend((f'E{edge.uid} to V{edge.v_i.uid}',
        #                    f'E{edge.uid} to V{edge.v_j.uid}', None))
        # vdf = pd.DataFrame(np.zeros((len(vertices), 2)))
        # vdf.columns = ['x', 'y']
        # vnames = []
        # for i, vertex_key in enumerate(vertices):
        #     vertex = vertices[vertex_key]
        #     vdf.loc[i, 'x':'y'] = vertex.coords
        #     vnames.append(vertex.uid)
        # edf = edf + np.random.normal(0.00, 0.10, edf.shape)
        # import plotly.express as px
        # import plotly.graph_objects as go
        # fig1 = px.line(edf, x='x', y='y', hover_name=enames)
        # fig2 = px.scatter(vdf, x='x', y='y', hover_name=vnames,
        #                   color=['red']*len(vnames))
        # fig = go.Figure(data=fig1.data + fig2.data)
        # fig.show()

        if perform_checks:
            # verify that no edges overlap
            all_edges_list = list(edges.values())
            for index, considered_edge in enumerate(
                tqdm(
                    all_edges_list, desc="Checking plan's edges", unit="edges"
                )
            ):
                if index == len(all_edges_list):
                    # we are done.
                    continue
                remaining_edges = all_edges_list[(index + 1)::]
                for other_edge in remaining_edges:
                    # check if the two edges overlap or cross each other
                    if considered_edge.overlaps_or_crosses(other_edge):
                        # the two edges overlap or cross each other
                        msg = "Error: Analysis of the floor plan geometry "
                        msg += "indicates the presence of "
                        msg += "overlapping elements.\n"
                        msg += "Check the model at the following locations:\n"
                        msg += f"{considered_edge.v_i.coords}"
                        msg += f"{considered_edge.v_j.coords}"
                        msg += f"{other_edge.v_i.coords}"
                        msg += f"{other_edge.v_j.coords}"
                        raise ValueError(msg)

        halfedges = mesh.define_halfedges(list(edges.values()))
        loops = mesh.obtain_closed_loops(halfedges)
        external, internal, trivial = mesh.orient_loops(loops)
        # Sanity checks.
        mesh.sanity_checks(external, trivial)

        for internal_loop in internal:
            poly = sg.Polygon([h.vertex.coords for h in internal_loop])
            skel = sg.skeleton.create_interior_straight_skeleton(poly)
            # todo: what we need to get rid of this is to end up with a list
            # of halfedges defining the subloops.

            # something like:
            # subloops = mesh.bisector_subdivision(internal_loop)

            subloops: list[list[Halfedge]] = []

            def is_in_some_subloop(halfedge, loops):
                for loop in loops:
                    for other_halfedge in loop:
                        if (
                            other_halfedge.vertex.point
                            == halfedge.vertex.point
                            and other_halfedge.next.vertex.point
                            == halfedge.next.vertex.point
                        ):
                            return True
                return False

            for halfedge in skel.halfedges:
                if subloops:
                    if is_in_some_subloop(halfedge, subloops):
                        continue
                subloop = [halfedge]
                nxt = halfedge.next
                while nxt.vertex.point != halfedge.vertex.point:
                    subloop.append(nxt)
                    nxt = nxt.next
                subloops.append(subloop)

            subloop_areas = [
                float(sg.Polygon([h.vertex.point for h in subloop]).area())
                for subloop in subloops
            ]
            outer = min(subloop_areas)  # Remove the exterior loop
            index = subloop_areas.index(outer)
            del subloops[index]
            del subloop_areas[index]

            for i, subloop in enumerate(subloops):
                area = subloop_areas[i]
                loop_edges = [h.edge for h in internal_loop]
                for halfedge in subloop:
                    for edge in loop_edges:
                        v_i = sg.Point2(*edge.v_i.coords)
                        v_j = sg.Point2(*edge.v_j.coords)
                        pt_1 = halfedge.vertex.point
                        pt_2 = halfedge.next.vertex.point
                        if (pt_1 == v_i and pt_2 == v_j) or (
                            pt_1 == v_j and pt_2 == v_i
                        ):
                            if edge.uid in edge_area:
                                edge_area[edge.uid] += area
                            else:
                                edge_area[edge.uid] = area
                            if edge.uid in edge_polygons:
                                edge_polygons[edge.uid].append(
                                    [
                                        (
                                            float(h.vertex.point.x()),
                                            float(h.vertex.point.y()),
                                        )
                                        for h in subloop
                                    ]
                                )
                            else:
                                edge_polygons[edge.uid] = [
                                    [
                                        (
                                            float(h.vertex.point.x()),
                                            float(h.vertex.point.y()),
                                        )
                                        for h in subloop
                                    ]
                                ]

        # # plotting - used while developing the code
        # import pandas as pd
        # import plotly.express as px
        # import plotly.graph_objects as go

        # x_vals = []
        # y_vals = []
        # colors = []
        # for edge in edges.values():
        #     x_vals.extend([edge.v_i.coords[0], edge.v_j.coords[0], None])
        #     y_vals.extend([edge.v_i.coords[1], edge.v_j.coords[1], None])
        #     colors.extend(['black', 'black', None])

        # fig1 = go.Figure(go.Scatter(x=x_vals, y=y_vals,
        #                             mode='lines',
        #                             line_color='black',
        #                             name='Elements'))

        # x_vals = []
        # y_vals = []
        # for edge in edges.values():
        #     for poly in edge_polygons[edge.uid]:
        #         poly_rot = poly[1:]+poly[:1]
        #         for vi, vj in zip(poly, poly_rot):
        #             x_vals.extend((vi[0], vj[0], None))
        #             y_vals.extend((vi[1], vj[1], None))

        # fig2 = go.Figure(go.Scatter(x=x_vals, y=y_vals,
        #                             mode='lines',
        #                             line_color='blue',
        #                             name='Regions'))

        # fig = go.Figure(data=fig2.data + fig1.data)
        # fig.update_yaxes(
        #     scaleanchor = "x",
        #     scaleratio = 1,
        #   )
        # fig.show()

        # apply loads and mass
        # todo (future): account for the shape of the loaded area
        mdl = self.parent_level.parent_model
        if mdl.settings.imperial_units:
            g_const = common.G_CONST_IMPERIAL
        else:
            g_const = common.G_CONST_SI
        lcase = self.parent_loadcase
        for uid, edge in edges.items():
            area = edge_area[uid]
            loaded_elm = edge_map[uid]
            for load in self.polygon_loads:
                if load.massless:
                    cmult = massless_load_factor
                else:
                    cmult = load_factor
                surf_load = load.value
                load_val = surf_load * area

                if isinstance(loaded_elm, Node):
                    lcase.node_loads[loaded_elm.uid].add(
                        np.array(
                            [0.00, 0.00, -load_val * cmult, 0.00, 0.00, 0.00]
                        )
                    )
                    if not load.massless:
                        lcase.node_mass[loaded_elm.uid].add(
                            np.array(
                                [
                                    load_val / g_const,
                                    load_val / g_const,
                                    load_val / g_const,
                                    0.00,
                                    0.00,
                                    0.00,
                                ]
                            )
                        )

                elif isinstance(
                    loaded_elm, (ElasticBeamColumn, DispBeamColumn)
                ):
                    length = loaded_elm.clear_length()
                    load_per_length = load_val / length
                    udl_obj = lcase.line_element_udl[loaded_elm.uid]
                    udl_obj.add_glob(
                        np.array([0.00, 0.00, -load_per_length * cmult])
                    )
                    if not load.massless:
                        half_mass = load_val / 2.00 / g_const
                        lcase.node_mass[loaded_elm.nodes[0].uid].add(
                            np.array([half_mass] * 3 + [0.00] * 3)
                        )
                        lcase.node_mass[loaded_elm.nodes[1].uid].add(
                            np.array([half_mass] * 3 + [0.00] * 3)
                        )

                elif isinstance(loaded_elm, TwoNodeLink):
                    for node in loaded_elm.nodes:
                        lcase.node_loads[node.uid].add(
                            np.array(
                                [
                                    0.00,
                                    0.00,
                                    -load_val * cmult / 2.00,
                                    0.00,
                                    0.00,
                                    0.00,
                                ]
                            )
                        )
                        if not load.massless:
                            lcase.node_mass[node.uid].add(
                                np.array(
                                    [
                                        load_val / g_const / 2.00,
                                        load_val / g_const / 2.00,
                                        load_val / g_const / 2.00,
                                        0.00,
                                        0.00,
                                        0.00,
                                    ]
                                )
                            )

                else:
                    raise TypeError("This should never happen!")



"""
Coordinate transformation operations.

"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

import numpy as np
import numpy.typing as npt
from . import common

nparr = npt.NDArray[np.float64]


def rotation_matrix_2d(ang: float) -> nparr:
    """
    Returns a 2D transformation matrix.

    Parameters:
        ang: Angle in radians to rotate the matrix by.

    Returns:
        A 2x2 transformation matrix.

    Example:
        >>> rotation_matrix_2d(np.pi / 2)
        array([[ 6.123234e-17, -1.000000e+00],
               [ 1.000000e+00,  6.123234e-17]])

    Raises:
        TypeError: If `ang` is not a float.

    """

    if not isinstance(ang, float):
        raise TypeError("ang parameter should be a float.")
    return np.array([[np.cos(ang), -np.sin(ang)], [np.sin(ang), np.cos(ang)]])


def rotation_matrix_3d(axis: nparr, theta: float) -> nparr:
    """
    Returns the rotation matrix associated with counterclockwise
    rotation about the given axis by theta radians.

    Parameters:
        axis: 3D vector representing the axis of rotation.
        theta: Angle of rotation in radians.

    Returns:
        3x3 transformation matrix representing the rotation.

    Example:
        >>> # this is how to run that function:
        >>> res = rotation_matrix_3d(np.array([1, 0, 0]), np.pi/2)
        >>> # this is the expected result:
        >>> expected_res = np.array(
        ...     [[ 1.00000000e+00,  0.00000000e+00, -0.00000000e+00],
        ...      [-0.00000000e+00,  2.22044605e-16, -1.00000000e+00],
        ...      [ 0.00000000e+00,  1.00000000e+00,  2.22044605e-16]])
        >>> assert np.allclose(res, expected_res)

    """

    v_a = np.cos(theta / 2.0)
    v_b, v_c, v_d = -axis * np.sin(theta / 2.0)
    v_aa, v_bb, v_cc, v_dd = v_a * v_a, v_b * v_b, v_c * v_c, v_d * v_d
    v_bc, v_ad, v_ac, v_ab, v_bd, v_cd = (
        v_b * v_c,
        v_a * v_d,
        v_a * v_c,
        v_a * v_b,
        v_b * v_d,
        v_c * v_d,
    )
    return np.array(
        [
            [v_aa + v_bb - v_cc - v_dd, 2 * (v_bc + v_ad), 2 * (v_bd - v_ac)],
            [2 * (v_bc - v_ad), v_aa + v_cc - v_bb - v_dd, 2 * (v_cd + v_ab)],
            [2 * (v_bd + v_ac), 2 * (v_cd - v_ab), v_aa + v_dd - v_bb - v_cc],
        ]
    )


def transformation_matrix(vec_x: nparr, vec_y: nparr, vec_z: nparr) -> nparr:
    """
    Returns a transformation matrix that transforms points from
    the coordinate system in which the x, y and z axes are expressed,
    to the local coordinate system defined by them.

    Arguments:
        vec_x: Local x axis expressed in the global system
        vec_y: (similar)
        vec_z: (similar)

    Returns:
        global to local transformation matrix.

    Note: For orthogonal axes, transpose to obtain the inverse transform.

    Example:
        >>> # this is how to run that function:
        >>> res = transformation_matrix(
        ...     np.array([1., 0., 0.]),
        ...     np.array([0., 1., 0.]),
        ...     np.array([0., 0., 1.]))
        >>> expected_result = np.array((
        ...     [[1., 0., 0.],
        ...      [0., 1., 0.],
        ...      [0., 0., 1.]]))
        >>> assert np.allclose(res, expected_result)
        >>> res = transformation_matrix(
        ...     np.array([1., 0., 0.]),
        ...     np.array([0., 0., 1.]),
        ...     np.array([0., 1., 0.]))
        >>> expected_result = np.array((
        ...     [[1., 0., 0.],
        ...      [0., 0., 1.],
        ...      [0., 1., 0.]]))
        >>> assert np.allclose(res, expected_result)

    """

    tr_global_to_local: nparr = np.vstack((vec_x, vec_y, vec_z))
    return tr_global_to_local


def local_axes_from_points_and_angle(
    point_i: nparr, point_j: nparr, ang: float
) -> tuple[nparr, nparr, nparr]:
    """
    Given a start point, and end point, and an angle,
    obtain the local coordinate system of a linear element.

    Arguments:
        point_i: Start point
        point_j: End point
        ang: Parameter that controls the rotation of the section
          around the x-axis. Counterclockwise rotation is
          posotive. 0.00 corresponds to vertical elements whose local
          z axis coincides with the local x axis, and horizontal
          elements whose local z axis is horizontal.

    Returns:
        Local coordinate system
        vectors. The first element is the local x axis, the second
        element is the local y axis, and the third element is the
        local z axis.

    Raises:
        ValueError: If the start point and end point define a vertical element
            that is defined upside down (i.e., with the start point at a lower
            height than the end point).

    Note:
        For vertical elements, the local x axis will be the vector connecting
        the start and end points, and the local z axis will be perpendicular
        to the local x axis and lying on the plane defined by the global xy
        plane and the local x axis. For horizontal elements, the local z axis
        will be parallel to the global z axis.

    Example:
        >>> point_i = np.array([0, 0, 0])
        >>> point_j = np.array([1, 0, 0])
        >>> ang = 0
        >>> local_axes_from_points_and_angle(point_i, point_j, ang)
        (array([1., 0., 0.]), array([0., 0., 1.]), array([ 0., -1.,  0.]))

    """

    # x-axis
    x_axis = point_j - point_i
    x_axis = x_axis / np.linalg.norm(x_axis)
    # y and z axes
    diff = np.abs(np.linalg.norm(x_axis - np.array([0.00, 0.00, -1.00])))
    if diff < common.EPSILON:
        # vertical case
        z_axis: nparr = np.array([np.cos(ang), np.sin(ang), 0.0])
        y_axis: nparr = np.cross(z_axis, x_axis)
    else:
        # not vertical case.
        # check if the element is upside down
        diff = np.abs(np.linalg.norm(x_axis - np.array([0.00, 0.00, 1.00])))
        if diff < common.EPSILON:
            raise ValueError("Vertical element defined upside down")
        up_direction: nparr = np.array([0.0, 0.0, 1.0])
        # orthogonalize with respect to x-axis
        y_axis = up_direction - np.dot(up_direction, x_axis) * x_axis
        # ..and normalize
        y_axis = y_axis / np.linalg.norm(y_axis)
        y_axis = np.dot(rotation_matrix_3d(x_axis, ang), y_axis)
        # z-axis
        z_axis = np.cross(x_axis, y_axis)

    return x_axis, y_axis, z_axis  # type: ignore


def offset_transformation(offset: nparr, u_vec: nparr, r_vec: nparr) -> nparr:
    """
    Calculate the displacement at the end of a rigid offset by
    specifying the displacement and rotation of the other end.

    A rigid offset connects two nodes and transmits forces between
    them, but does not allow any relative displacement or rotation
    between the nodes.

    Arguments:
        offset:
          Vector pointing from the node of the rigid offset where the
          displacement is known to the node where we want to obtain
          the displacement. The vector should be given in the global
          coordinate system.
        u_vec:
          Displacement of the node where the displacement is known,
          given in the global coordinate system.
        r_vec:
          Rotation of the node where the displacement is known, given
          as a vector of the form [rx, ry, rz] representing the
          rotation around the x, y, and z axes, respectively.

    Returns:
        Displacement at the other end of the rigid offset,
        given in the global coordinate system.

    Example:
        Calculate the displacement of the end of a rigid offset with a
        length of 1 meter, given a displacement of [4, 5, 6] and a
        rotation of [7, 8, 9] at the other end:

        >>> offset_transformation(np.array([1., 0., 0.]),
        ...     np.array([0.01, -0.02, 0.005]),
        ...     np.array([0.0002, -0.0003, 0.0001]))
        array([ 0.01  , -0.0199,  0.0053])

    """

    t_rigid: nparr = np.array(
        [
            [0.00, +offset[2], -offset[1]],
            [-offset[2], 0.00, +offset[0]],
            [+offset[1], -offset[0], 0.00],
        ]
    )
    return u_vec + t_rigid @ r_vec



"""
Objects that generate sections.

"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from typing import TYPE_CHECKING
from typing import Type
from dataclasses import dataclass
import json
import pkgutil
import numpy as np
import numpy.typing as npt
from ..ops.section import SectionComponent
from ..ops.section import ElasticSection
from ..ops.section import FiberSection
from ..gen import mesh_shapes
from .mesh_shapes import rect_mesh

if TYPE_CHECKING:
    from ..model import Model
    from ..ops.section import Section

nparr = npt.NDArray[np.float64]


@dataclass(repr=False)
class SectionGenerator:
    """
    Used to populate the section repository of a model.

    Attributes:
      model: Model to act upon

    """

    model: Model

    def generate_generic_elastic(
        self, name: str, e_times_a: float, e_times_i: float, g_times_j: float
    ) -> ElasticSection:
        """
        Generates an ElasticSection object with the specified properties.

        Arguments:
            name: Name of the section.
            e_times_a:
                The product of the elastic modulus and the area of the
                section.
            e_times_i:
                The product of the elastic modulus and the moment of
                inertia.
            g_times_j:
                The product of the shear modulus and the torsional
                moment of inertia.

        Returns:
            ElasticSection: An ElasticSection object with the
                specified properties.

        Example:
            >>> from osmg.gen.section_gen import SectionGenerator
            >>> from osmg.model import Model
            >>> model = Model('test_model')
            >>> sec_gen = SectionGenerator(model)
            >>> sec = sec_gen.generate_generic_elastic(
            ...     name="My Elastic Section",
            ...     e_times_a=100.00, e_times_i=1000.00, g_times_j=500.00)
            >>> sec.name
            'My Elastic Section'
            >>> sec.area
            100.0
            >>> sec.i_y
            1000.0
            >>> sec.i_x
            1000.0
            >>> sec.j_mod
            500.0

        """

        sec = ElasticSection(
            name=name,
            uid=self.model.uid_generator.new("section"),
            outside_shape=None,
            snap_points=None,
            e_mod=1.00,
            area=e_times_a,
            i_y=e_times_i,
            i_x=e_times_i,
            g_mod=1.00,
            j_mod=g_times_j,
            sec_w=0.00,
        )

        if self.model.settings.imperial_units:
            y_max = +10.00
            y_min = -10.00
            z_max = +6.00
            z_min = -6.00
            sec.outside_shape = rect_mesh(12.0, 20.0)
        else:
            y_max = +0.25
            y_min = -0.25
            z_max = +0.15
            z_min = -0.15
            sec.outside_shape = rect_mesh(0.30, 0.50)
        snap_points: dict[str, nparr] = {
            "centroid": np.array([0.0, 0.0]),
            "top_center": np.array([0.0, -y_max]),
            "top_left": np.array([-z_min, -y_max]),
            "top_right": np.array([-z_max, -y_max]),
            "center_left": np.array([-z_min, 0.0]),
            "center_right": np.array([-z_max, 0.0]),
            "bottom_center": np.array([0.0, -y_min]),
            "bottom_left": np.array([-z_min, -y_min]),
            "bottom_right": np.array([-z_max, -y_min]),
        }
        sec.snap_points = snap_points
        self.model.elastic_sections.add(sec)
        return sec

    def load_aisc_from_database(
        self,
        sec_shape_designation: str,
        labels: list[str],
        ops_material: str,
        physical_material: str,
        sec_type: Type[Section],
        store_in_model: bool = True,
        return_section: bool = False,
    ) -> dict[str, ElasticSection | FiberSection]:
        """
        Loads a section from the AISC steel section database.

        Arguments:
            sec_shape_designation: Designation of the type of section
                to be loaded.
            labels: List of labels of the sections to be loaded.
            ops_material: Name of the uniaxial material to be
                associated with the section.
            physical_material: Name of the physical material to be
                associated with the section.
            sec_type: Type of section to be created.
            store_in_model: If True, the created sections are stored
                in the model.
            return_section: If True, the created sections are returned.

        Returns:
            If `return_section` is True, a dictionary containing the
            created sections. The keys are the labels of the sections,
            and the values are the sections themselves.

        Example:
            >>> from osmg.gen.section_gen import SectionGenerator
            >>> from osmg.model import Model
            >>> from osmg.defaults import load_default_steel
            >>> from osmg.ops.section import ElasticSection
            >>> model = Model('test_model')
            >>> load_default_steel(model)
            >>> sec_gen = SectionGenerator(model)
            >>> sec_gen.load_aisc_from_database(
            ...     'W', ['W14X90'], 'default steel', 'default steel',
            ...     ElasticSection, store_in_model=True, return_section=True)
            {'W14X90': ElasticSection object
            name: W14X90
            uid: 0
            Properties:  E: 29000000.0
              A: 26.5
              Iy: 362.0
              Ix: 999.0
              G: 11153846.15
              J: 4.06
              W: 7.5
            outside_shape: specified
            snap_points: specified
            }

        """

        ops_mat = self.model.uniaxial_materials.retrieve_by_attr(
            "name", ops_material
        )
        phs_mat = self.model.physical_materials.retrieve_by_attr(
            "name", physical_material
        )
        filename = "../sections.json"
        contents = pkgutil.get_data(__name__, filename)
        assert isinstance(contents, bytes)
        section_dictionary = json.loads(contents)
        assert self.model.settings.imperial_units, "SI not supported"
        returned_sections: dict[str, ElasticSection | FiberSection] = {}
        for label in labels:
            try:
                sec_data = section_dictionary[label]
            except KeyError as exc:
                raise KeyError(f"Section {label} not found in file.") from exc
            if sec_shape_designation == "W":
                assert sec_data["Type"] == "W"
                sec_b = sec_data["bf"]
                sec_h = sec_data["d"]
                sec_tw = sec_data["tw"]
                sec_tf = sec_data["tf"]
                area = sec_data["A"]
                outside_shape = mesh_shapes.w_mesh(
                    sec_b, sec_h, sec_tw, sec_tf, area
                )
                bbox = outside_shape.bounding_box()
                z_min, y_min, z_max, y_max = bbox.flatten()
                snap_points: dict[str, nparr] = {
                    "centroid": np.array([0.0, 0.0]),
                    "top_center": np.array([0.0, -y_max]),
                    "top_left": np.array([-z_min, -y_max]),
                    "top_right": np.array([-z_max, -y_max]),
                    "center_left": np.array([-z_min, 0.0]),
                    "center_right": np.array([-z_max, 0.0]),
                    "bottom_center": np.array([0.0, -y_min]),
                    "bottom_left": np.array([-z_min, -y_min]),
                    "bottom_right": np.array([-z_max, -y_min]),
                }
                if sec_type.__name__ == "FiberSection":
                    main_part = SectionComponent(
                        outside_shape, {}, ops_mat, phs_mat
                    )
                    sec_fib = FiberSection(
                        name=label,
                        uid=self.model.uid_generator.new("section"),
                        outside_shape=outside_shape,
                        section_parts={"main": main_part},
                        j_mod=sec_data["J"],
                        snap_points=snap_points,
                        properties=sec_data,
                        n_x=10,
                        n_y=10,
                    )
                    if store_in_model:
                        self.model.fiber_sections.add(sec_fib)
                    if return_section:
                        returned_sections[sec_fib.name] = sec_fib
                elif sec_type.__name__ == "ElasticSection":
                    sec_el = ElasticSection(
                        label,
                        self.model.uid_generator.new("section"),
                        phs_mat.e_mod,
                        sec_data["A"],
                        sec_data["Iy"],
                        sec_data["Ix"],
                        phs_mat.g_mod,
                        sec_data["J"],
                        sec_data["W"] / 12.00,  # lb/in
                        outside_shape,
                        snap_points,
                        properties=sec_data,
                    )
                    if store_in_model:
                        self.model.elastic_sections.add(sec_el)
                    if return_section:
                        returned_sections[sec_el.name] = sec_el
                else:
                    raise ValueError(
                        f"Unsupported section type: {sec_type.__name__}"
                    )
            elif sec_shape_designation == "HSS_rect":
                assert sec_data["Type"] == "HSS"
                # must be rectangle: name will have 2 X's.
                assert len(label.split("X")) == 3
                sec_ht = sec_data["Ht"]
                sec_b = sec_data["B"]
                sec_t = sec_data["tdes"]
                outside_shape = mesh_shapes.rect_mesh(sec_b, sec_ht)
                hole = mesh_shapes.rect_mesh(
                    sec_b - 2.00 * sec_t, sec_ht - 2.00 * sec_t
                )
                bbox = outside_shape.bounding_box()
                z_min, y_min, z_max, y_max = bbox.flatten()
                snap_points = {
                    "centroid": np.array([0.0, 0.0]),
                    "top_center": np.array([0.0, -y_max]),
                    "top_left": np.array([-z_min, -y_max]),
                    "top_right": np.array([-z_max, -y_max]),
                    "center_left": np.array([-z_min, 0.0]),
                    "center_right": np.array([-z_max, 0.0]),
                    "bottom_center": np.array([0.0, -y_min]),
                    "bottom_left": np.array([-z_min, -y_min]),
                    "bottom_right": np.array([-z_max, -y_min]),
                }
                if sec_type.__name__ == "FiberSection":
                    main_part = SectionComponent(
                        outside_shape, {"hole": hole}, ops_mat, phs_mat
                    )
                    sec_fib = FiberSection(
                        label,
                        self.model.uid_generator.new("section"),
                        outside_shape,
                        {"main": main_part},
                        sec_data["J"],
                        snap_points,
                        sec_data,
                        n_x=10,
                        n_y=10,
                    )
                    if store_in_model:
                        self.model.fiber_sections.add(sec_fib)
                    if return_section:
                        returned_sections[sec_fib.name] = sec_fib
                elif sec_type.__name__ == "ElasticSection":
                    sec_el = ElasticSection(
                        label,
                        self.model.uid_generator.new("section"),
                        phs_mat.e_mod,
                        sec_data["A"],
                        sec_data["Iy"],
                        sec_data["Ix"],
                        phs_mat.g_mod,
                        sec_data["J"],
                        sec_data["W"] / 12.00,  # lb/in
                        outside_shape,
                        snap_points,
                        properties=sec_data,
                    )
                    if store_in_model:
                        self.model.elastic_sections.add(sec_el)
                    if return_section:
                        returned_sections[sec_el.name] = sec_el
                else:
                    raise ValueError(
                        f"Unsupported section type: {sec_type.__name__}"
                    )
            elif sec_shape_designation == "HSS_circ":
                # TODO: eliminate some redundant code here by merging
                # suare and round HSS
                assert sec_data["Type"] == "HSS"
                # must be circular: name will have 1.
                assert len(label.split("X")) == 2
                sec_h = sec_data["OD"]
                sec_t = sec_data["tdes"]
                outside_shape = mesh_shapes.circ_mesh(sec_h)
                hole = mesh_shapes.circ_mesh(sec_h - 2.00 * sec_t)
                bbox = outside_shape.bounding_box()
                z_min, y_min, z_max, y_max = bbox.flatten()
                snap_points = {
                    "centroid": np.array([0.0, 0.0]),
                    "top_center": np.array([0.0, -y_max]),
                    "top_left": np.array([-z_min, -y_max]),
                    "top_right": np.array([-z_max, -y_max]),
                    "center_left": np.array([-z_min, 0.0]),
                    "center_right": np.array([-z_max, 0.0]),
                    "bottom_center": np.array([0.0, -y_min]),
                    "bottom_left": np.array([-z_min, -y_min]),
                    "bottom_right": np.array([-z_max, -y_min]),
                }
                if sec_type.__name__ == "FiberSection":
                    main_part = SectionComponent(
                        outside_shape, {"hole": hole}, ops_mat, phs_mat
                    )
                    sec_fib = FiberSection(
                        label,
                        self.model.uid_generator.new("section"),
                        outside_shape,
                        {"main": main_part},
                        sec_data["J"],
                        snap_points,
                        sec_data,
                        n_x=10,
                        n_y=10,
                    )
                    if store_in_model:
                        self.model.fiber_sections.add(sec_fib)
                    if return_section:
                        returned_sections[sec_fib.name] = sec_fib
                elif sec_type.__name__ == "ElasticSection":
                    sec_el = ElasticSection(
                        label,
                        self.model.uid_generator.new("section"),
                        phs_mat.e_mod,
                        sec_data["A"],
                        sec_data["Iy"],
                        sec_data["Ix"],
                        phs_mat.g_mod,
                        sec_data["J"],
                        sec_data["W"] / 12.00,  # lb/in
                        outside_shape,
                        snap_points,
                        properties=sec_data,
                    )
                    if store_in_model:
                        self.model.elastic_sections.add(sec_el)
                    if return_section:
                        returned_sections[sec_el.name] = sec_el
                else:
                    raise ValueError(
                        f"Unsupported section type: {sec_type.__name__}"
                    )
            else:
                raise ValueError(
                    "Unsupported section designtation:"
                    f" {sec_shape_designation}"
                )
        return returned_sections



"""
Defines Model objects.

"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

# pylint: disable=W1512

from __future__ import annotations
from typing import TYPE_CHECKING
from typing import Type
from dataclasses import dataclass, field
import numpy as np
import numpy.typing as npt
from shapely.geometry import Polygon as shapely_Polygon  # type: ignore
from shapely.geometry import Point  # type: ignore

from .gen.uid_gen import UIDGenerator
from . import obj_collections
from .level import Level

if TYPE_CHECKING:
    from .ops.node import Node
    from .ops import element
    from .component_assembly import ComponentAssembly
    from .ops.section import ElasticSection
    from .ops.section import FiberSection
    from .ops.uniaxial_material import UniaxialMaterial
    from .physical_material import PhysicalMaterial

nparr = npt.NDArray[np.float64]


def transfer_component(other: Model, component: ComponentAssembly) -> None:
    """
    Transfers a single component assembly from one model to
    another, assuming the other model was generated with the
    :func:`~Model.initialize_empty_copy` method.

    """

    # note: we don't copy the component assemblies and their contents.
    # we just add the same objects to the other model.
    level = component.parent_collection.parent
    other_level = other.levels.retrieve_by_attr("uid", level.uid)
    for node in component.external_nodes.values():
        if node.uid not in other_level.nodes:
            other_level.nodes.add(node)
    other_level.components.add(component)


@dataclass
class Settings:
    """
    General customization of a model.
        imperial_units:
            True for imperial:
                in, lb, lb/(in/s2)
            False for SI:
                m, N, kg
        ndm, ndf: Number of dimensions and degrees of freedom.
          Currently only ndm=3, ndf=6 is supported. This might be
          extended in the future if needed.

    """

    imperial_units: bool = field(default=True)  # false for SI
    ndm: int = field(default=3)  # that's all we support
    ndf: int = field(default=6)  # that's all we support

    def __repr__(self):
        res = ""
        res += "~~~ Model Settings ~~~\n"
        res += f"  Imperial units: {self.imperial_units}\n"
        res += f"  ndm           : {self.ndm}\n"
        res += f"  ndf           : {self.ndf}\n"
        return res


@dataclass(repr=False)
class Model:
    """
    Model object.
    A model object is a representation of a structural model in
    OpenSees. It is the primary object of osmg.
    It contains levels, elastic sections, fiber sections,
    physical materials, and various collections of objects such as
    nodes, elements, and component assemblies that exist inside each level.
    Those objects are populated by generator objects. See `osmg.gen`.

    Attributes:
        name: Name of the model.
        levels:
            Collection of levels in the model.
        elastic_sections:
            Collection of elastic sections in the model.
        fiber_sections:
            Collection of fiber sections in the model.
        uniaxial_materials:
            Collection of uniaxial materials in the model.
        physical_materials:
            Collection of physical materials in the model.
        uid_generator:
            Object for generating unique IDs for objects in the model.
        settings: Settings for the model.

    """

    name: str
    levels: obj_collections.CollectionActive[int, Level] = field(init=False)
    elastic_sections: obj_collections.Collection[int, ElasticSection] = field(
        init=False
    )
    fiber_sections: obj_collections.Collection[int, FiberSection] = field(
        init=False
    )
    uniaxial_materials: obj_collections.Collection[
        int, UniaxialMaterial
    ] = field(init=False)
    physical_materials: obj_collections.Collection[
        int, PhysicalMaterial
    ] = field(init=False)
    uid_generator: UIDGenerator = field(default_factory=UIDGenerator)
    settings: Settings = field(default_factory=Settings)

    def __post_init__(self):
        self.levels = obj_collections.CollectionActive(self)
        self.elastic_sections = obj_collections.Collection(self)
        self.fiber_sections = obj_collections.Collection(self)
        self.uniaxial_materials = obj_collections.Collection(self)
        self.physical_materials = obj_collections.Collection(self)

    def __repr__(self):
        res = ""
        res += "~~~ Model Object ~~~\n"
        res += f"ID: {id(self)}\n"
        res += f"levels: {self.levels.__srepr__()}\n"
        res += f"elastic_sections: {self.elastic_sections.__srepr__()}\n"
        res += f"fiber_sections: {self.fiber_sections.__srepr__()}\n"
        res += f"uniaxial_materials: {self.uniaxial_materials.__srepr__()}\n"
        res += f"physical_materials: {self.physical_materials.__srepr__()}\n"
        return res

    def component_connectivity(
        self,
    ) -> dict[tuple[int, ...], ComponentAssembly]:
        """
        Returns the connectivity of all component
        assemblies. Component assemblies are collections of
        lower-level components that are connected to primary
        nodes. Each component assembly can be represented by a tuple
        of node uids of its connected nodes in ascending order. This
        method returns a dictionary having these tuples as keys, and
        the associated components as values.

        """

        res = {}
        components = self.list_of_components()
        for component in components:
            uids = [node.uid for node in component.external_nodes.values()]
            uids.sort()
            uids_tuple = (*uids,)
            assert uids_tuple not in res, "Error! Duplicate component found."
            res[uids_tuple] = component
        return res

    def add_level(self, uid: int, elevation: float) -> None:
        """
        Adds a level to the model.

        Arguments:
            uid: Unique ID for the level.
            elevation: Elevation of the level.

        Example:
            >>> from osmg.model import Model
            >>> model = Model('test_model')
            >>> model.add_level(1, 0.0)
            >>> model.levels.__srepr__()
            '[Collection of 1 items]'

        """

        lvl = Level(self, uid=uid, elevation=elevation)
        self.levels.add(lvl)

    def dict_of_primary_nodes(self):
        """
        Returns a dictionary of all the primary nodes in the model.
        The keys are the uids of the nodes.
        """
        dict_of_nodes: dict[int, Node] = {}
        for lvl in self.levels.values():
            dict_of_nodes.update(lvl.nodes)
        return dict_of_nodes

    def list_of_primary_nodes(self):
        """
        Returns a list of all the primary nodes in the model.

        """

        list_of_nodes = []
        for lvl in self.levels.values():
            for node in lvl.nodes.values():
                list_of_nodes.append(node)
        return list_of_nodes

    def dict_of_internal_nodes(self):
        """
        Returns a dictionary of all the internal nodes in the model.
        The keys are the uids of the nodes.

        """

        dict_of_nodes: dict[int, Node] = {}
        for lvl in self.levels.values():
            for component in lvl.components.values():
                dict_of_nodes.update(component.internal_nodes)
        return dict_of_nodes

    def list_of_internal_nodes(self):
        """
        Returns a list of all the internal nodes in the model.

        """

        list_of_nodes = []
        for lvl in self.levels.values():
            for component in lvl.components.values():
                for inode in component.internal_nodes.values():
                    list_of_nodes.append(inode)
        return list_of_nodes

    def dict_of_all_nodes(self):
        """
        Returns a dictionary of all the nodes in the model.
        The keys are the uids of the nodes.

        """

        dict_of_nodes: dict[int, Node] = {}
        dict_of_nodes.update(self.dict_of_primary_nodes())
        dict_of_nodes.update(self.dict_of_internal_nodes())
        return dict_of_nodes

    def list_of_all_nodes(self):
        """
        Returns a list of all the nodes in the model.

        """

        list_of_nodes = []
        list_of_nodes.extend(self.list_of_primary_nodes())
        list_of_nodes.extend(self.list_of_internal_nodes())
        return list_of_nodes

    def dict_of_components(self):
        """
        Returns a dictionary of all the component assemblies in the
        model.
        The keys are the uids of the component assemblies.

        """

        comps: dict[int, ComponentAssembly] = {}
        for lvl in self.levels.values():
            for component in lvl.components.values():
                comps[component.uid] = component
        return comps

    def list_of_components(self):
        """
        Returns a list of all the component assembiles in the
        model.

        """

        return list(self.dict_of_components().values())

    def dict_of_elements(self) -> dict[int, element.Element]:
        """
        Returns a dictionary of all element objects in the model.
        The keys are the uids of the objects.

        """

        elems: dict[int, element.Element] = {}
        for lvl in self.levels.values():
            for component in lvl.components.values():
                elems.update(component.elements)  # type: ignore
        return elems

    def list_of_elements(self) -> list[element.Element]:
        """
        Returns a list of all element objects in the model.

        """

        return list(self.dict_of_elements().values())

    def dict_of_specific_element(
        self, element_class: Type[element.Element]
    ) -> dict[int, element.Element]:
        """
        Returns a dictionary of all element objects in the model of a
        particular element class.
        The keys are the uids of the objects.

        """

        all_elements = self.dict_of_elements()
        res: dict[int, element.Element] = {}
        for uid, elm in all_elements.items():
            if isinstance(elm, element_class):
                res[uid] = elm
        return res

    def list_of_specific_element(
        self, element_class: Type[element.Element]
    ) -> list[element.Element]:
        """
        Returns a list of all element objects in the model of a
        particular element class.

        """

        return list(self.dict_of_specific_element(element_class).values())

    def bounding_box(self, padding: float) -> tuple[nparr, nparr]:
        """
        Returns the axis-aligned bouding box of the building

        """

        p_min = np.full(3, np.inf)
        p_max = np.full(3, -np.inf)
        for node in self.list_of_primary_nodes():
            point: nparr = np.array(node.coords)
            p_min = np.minimum(p_min, point)
            p_max = np.maximum(p_max, point)
        p_min -= np.full(3, padding)
        p_max += np.full(3, padding)
        # type hints gone mad  >.<   ...
        return p_min, p_max  # type:ignore

    def reference_length(self):
        """
        Returns the largest dimension of the
        bounding box of the building
        (used in graphics)

        """

        p_min, p_max = self.bounding_box(padding=0.00)
        ref_len = np.max(p_max - p_min)
        return ref_len

    def initialize_empty_copy(self, name):
        """
        Initializes a shallow empty copy of the model.
        Used to create subset models.

        """

        res = Model(name)
        # copy the settings attributes
        res.settings.imperial_units = self.settings.imperial_units
        res.settings.ndf = self.settings.ndf
        res.settings.ndm = self.settings.ndm
        # make a copy of the levels
        for lvlkey, lvl in self.levels.items():
            res.add_level(lvlkey, lvl.elevation)
        # giv access to the materials and sections
        res.elastic_sections = self.elastic_sections
        res.fiber_sections = self.fiber_sections
        res.physical_materials = self.physical_materials
        res.uniaxial_materials = self.uniaxial_materials
        return res

    def transfer_by_polygon_selection(
        self, other: Model, coords: nparr
    ) -> None:
        """
        Uses :func:`~transfer_component` to transfer all components of which
        the projection to the XY plane falls inside the specified
        polygon.

        """

        all_components = self.list_of_components()
        selected_components = []
        shape = shapely_Polygon(coords)
        for component in all_components:
            accept = True
            nodes = component.external_nodes.values()
            for node in nodes:
                if not shape.contains(Point(node.coords[0:2])):
                    accept = False
                    break
            if accept:
                selected_components.append(component)
        for component in selected_components:
            transfer_component(other, component)



"""
Defines :obj:`~osmg.ops.section.Section` objects.
"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from typing import TYPE_CHECKING
from typing import Optional
from typing import Any
from dataclasses import dataclass, field
import numpy as np
import numpy.typing as npt
from ..mesh import Mesh
from ..mesh import polygon_area
from .. import common
from .uniaxial_material import UniaxialMaterial
from .. import mesh

if TYPE_CHECKING:
    from ..physical_material import PhysicalMaterial

nparr = npt.NDArray[np.float64]


@dataclass
class Section:
    """
    Section object.
    The axes are defined in the same way as they are
    defined in OpenSees. The colors assigned to
    the axes for plotting follow the
    AutoCAD convention.

    .. code-block:: python

            y(green)
            ^         x(red)
            :       .
            :     .
            :   .
           ===
            | -------> z (blue)
           ===

    """

    name: str
    uid: int


@dataclass
class ElasticSection(Section):
    """
    Elastic Section Object.

    Attributes:
      e_mod: Young's modulus.
      area: Cross-sectional area.
      i_y: Moment of inertia for strong-axis bending.
      i_x: Moment of inertia for weak-axis bending.
      g_mod: Shear modulus.
      j_mod: Torsional moment of inertia.
      sec_w: Weight per unit length.
      outside_shape: Mesh defining the outside shape of the section.
      snap_points: Dictionary containing coordinates of `snap_points`
        used by component-generating methods to position components
        relative to existing ones. See
        :func:`~osmg.gen.component_gen.beam_placement_lookup` for example.
      properties: Dictionary containing section properties.

    """

    e_mod: float
    area: float
    i_y: float
    i_x: float
    g_mod: float
    j_mod: float
    sec_w: float
    outside_shape: Optional[Mesh] = field(default=None, repr=False)
    snap_points: Optional[dict[str, nparr]] = field(default=None, repr=False)
    properties: Optional[dict[str, Any]] = field(default=None, repr=False)

    def weight_per_length(self):
        """
        Returns the weight per length of a section.
        For steel W sections, it adds 15% for misc. steel and
        connections.

        """

        if self.name[0] == "W":
            res = self.sec_w * 1.15  # misc steel and connections
        else:
            res = self.sec_w
        return res

    def __repr__(self):
        res = ""
        res += "ElasticSection object\n"
        res += f"name: {self.name}\n"
        res += f"uid: {self.uid}\n"
        res += "Properties:"
        res += f"  E: {self.e_mod}\n"
        res += f"  A: {self.area}\n"
        res += f"  Iy: {self.i_y}\n"
        res += f"  Ix: {self.i_x}\n"
        res += f"  G: {self.g_mod}\n"
        res += f"  J: {self.j_mod}\n"
        res += f"  W: {self.sec_w}\n"
        if self.outside_shape:
            res += "outside_shape: specified\n"
        else:
            res += "outside_shape: None\n"
        if self.snap_points:
            res += "snap_points: specified\n"
        else:
            res += "snap_points: None\n"
        return res


@dataclass(repr=False)
class SectionComponent:
    """
    Part of a section object, having a single material.

    Arguments:
      outside_shape: Mesh defining the outside shape
      ops_material: OpenSees material
      physical_material: Physical material
      parent_section: Parent section.
        The parent section is assigned automatically by their
        parent section iteslf, at its creation time.

    """

    outside_shape: Mesh
    holes: dict[str, Mesh]
    ops_material: UniaxialMaterial
    physical_material: PhysicalMaterial
    parent_section: Optional[FiberSection] = field(default=None)

    def __repr__(self):
        res = ""
        res += "SectionComponent object\n"
        if self.outside_shape:
            res += "outside_shape: specified\n"
        else:
            res += "outside_shape: None\n"
        if self.holes:
            res += "holes: exist\n"
        else:
            res += "holes: no holes\n"
        res += f"ops_material: {self.ops_material.name}\n"
        res += f"physical_material: {self.physical_material.name}\n"
        return res

    def cut_into_tiny_little_pieces(self):
        """
        Returns data used to define fibers in OpenSees.

        """

        # if we have an AISC HSS section, we need to discretize in a
        # certain way
        assert self.parent_section
        sec_name = self.parent_section.name
        if "HSS" in sec_name and len(sec_name.split("X")) == 3:
            # rectangular HSS section!
            pieces = mesh.subdivide_hss_rect(
                self.parent_section.properties["Ht"],
                self.parent_section.properties["B"],
                self.parent_section.properties["tdes"],
            )
        elif 'HSS' in sec_name and len(sec_name.split('X')) == 2:
            # circular HSS section!
            pieces = mesh.subdivide_hss_circ(
                self.parent_section.properties["OD"],
                self.parent_section.properties["tdes"],
            )

        else:
            # fallback: use the default rectangular mesh chopper
            pieces = mesh.subdivide_polygon(
                self.outside_shape,
                self.holes,
                self.parent_section.n_x,
                self.parent_section.n_y,
            )

        return pieces

    def copy_alter_material(self, mat: UniaxialMaterial) -> SectionComponent:
        """
        Make a shallow copy of a section component and replace the old
        material with the given one.

        """

        new_component = SectionComponent(
            self.outside_shape, self.holes, mat, self.physical_material
        )
        return new_component


@dataclass(repr=False)
class FiberSection(Section):
    """
    Fiber section object.
    Can consist of multiple materials.
    The primary part of the component must have the key "main".

    """

    outside_shape: Mesh
    section_parts: dict[str, SectionComponent]
    j_mod: float
    snap_points: dict[str, nparr]
    properties: dict[str, Any]
    n_x: int
    n_y: int

    def __post_init__(self):
        for part in self.section_parts:
            self.section_parts[part].parent_section = self

    def __repr__(self):
        res = ""
        res += "FiberSection object\n"
        for part in self.section_parts:
            res += part.__repr__()
        if self.snap_points:
            res += "snap_points: specified\n"
        else:
            res += "snap_points: None\n"
        res += f"n_x: {self.n_x}, n_y: {self.n_y}\n"
        return res

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees

        """

        return [
            "Fiber",
            self.uid,
            "-GJ",
            self.j_mod * self.section_parts["main"].physical_material.g_mod,
        ]

    def weight_per_length(self):
        """
        Returns the weight per length of a section.
        For steel W sections, it adds 15% for misc. steel and connections.

        """

        if self.name[0] == "W":
            mult = 1.15  # misc steel and connections
        else:
            mult = 1.00
        res = 0.00
        for part in self.section_parts.values():
            coords: nparr = np.array(
                [h.vertex.coords for h in part.outside_shape.halfedges]
            )
            area = polygon_area(coords)
            for hole in part.holes:
                hole_coords: nparr = np.array(
                    [h.vertex.coords for h in part.holes[hole].halfedges]
                )
                area -= polygon_area(hole_coords)
            density = part.physical_material.density
            # TODO: units
            res += area * density * common.G_CONST_IMPERIAL
        return res * mult

    def copy_alter_material(
        self, mat: UniaxialMaterial, new_uid: int
    ) -> FiberSection:
        """
        Returns a shallow copy of the section object in which all
        opensees_material objects have been replaced with the given
        material. Required for the modling of steel braced frames.

        """

        new_section_parts = {}
        for key, val in self.section_parts.items():
            new_part = val.copy_alter_material(mat)
            new_section_parts[key] = new_part
        other_sec = FiberSection(
            f"auto_{self.name}",
            new_uid,
            self.outside_shape,
            new_section_parts,
            self.j_mod,
            self.snap_points,
            self.properties,
            self.n_x,
            self.n_y,
        )
        return other_sec



"""
Defines OpenSees uniaxialMaterial interfrace objects.

"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from typing import Optional
from dataclasses import dataclass
from dataclasses import field


# pylint: disable=invalid-name


@dataclass
class UniaxialMaterial:
    """
    OpenSees uniaxialMaterial
    https://openseespydoc.readthedocs.io/en/latest/src/uniaxialMaterial.html

    """

    uid: int
    name: str


@dataclass
class Elastic(UniaxialMaterial):
    """
    OpenSees Elastic
    https://openseespydoc.readthedocs.io/en/latest/src/ElasticUni.html

    """

    e_mod: float

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees
        """
        return ["Elastic", self.uid, self.e_mod]


@dataclass
class Steel02(UniaxialMaterial):
    """
    OpenSees Steel02
    https://openseespydoc.readthedocs.io/en/latest/src/steel02.html

    """

    Fy: float
    E0: float
    G: float
    b: float
    c_r0: float
    c_r1: float
    c_r2: float
    a1: Optional[float] = field(default=None)
    a2: Optional[float] = field(default=None)
    a3: Optional[float] = field(default=None)
    a4: Optional[float] = field(default=None)
    sig_init: Optional[float] = field(default=None)

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees

        """

        args = [
            "Steel02",
            self.uid,
            self.Fy,
            self.E0,
            self.b,
            self.c_r0,
            self.c_r1,
            self.c_r2,
        ]
        if self.a1:
            args.append(self.a1)
        if self.a2:
            args.append(self.a2)
        if self.a3:
            args.append(self.a3)
        if self.a4:
            args.append(self.a4)
        if self.sig_init:
            args.append(self.sig_init)

        return args


@dataclass
class Steel4(UniaxialMaterial):
    """
    OpenSees Steel4
    https://openseespydoc.readthedocs.io/en/latest/src/steel4.html

    """

    Fy: float
    E0: float
    b_k: Optional[float] = field(default=None)
    R_0: float = field(default=20.00)
    r_1: float = field(default=0.90)
    r_2: float = field(default=0.15)

    b_kc: Optional[float] = field(default=False)
    R_0c: float = field(default=20.00)
    r_1c: float = field(default=0.90)
    r_2c: float = field(default=0.15)

    b_i: Optional[float] = field(default=None)
    b_l: Optional[float] = field(default=None)
    rho_i: Optional[float] = field(default=None)
    R_i: Optional[float] = field(default=None)
    l_yp: Optional[float] = field(default=None)
    f_u: Optional[float] = field(default=None)
    R_u: Optional[float] = field(default=None)

    f_uc: Optional[float] = field(default=None)
    R_uc: Optional[float] = field(default=None)
    b_ic: Optional[float] = field(default=None)
    b_lc: Optional[float] = field(default=None)
    rho_ic: Optional[float] = field(default=None)
    R_ic: Optional[float] = field(default=None)

    sig_init: Optional[float] = field(default=None)
    cycNum: Optional[float] = field(default=None)

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees

        """

        # non-symmetric behavior
        if self.b_kc:
            assert self.R_0c is not None
            assert self.r_1c is not None
            assert self.r_2c is not None
            assert self.b_k
            asym = True
        else:
            asym = False
        # ultimate strength limit
        if self.f_u:
            ultimate = True
            if asym:
                assert self.f_uc is not None
                assert self.R_uc is not None
        else:
            ultimate = False
        # isotropic hardening
        if self.b_i:
            iso = True
            assert self.b_l is not None
            assert self.rho_i is not None
            assert self.R_i is not None
            assert self.l_yp is not None
            if asym:
                assert self.b_lc is not None
                assert self.rho_ic is not None
                assert self.R_ic is not None
        else:
            iso = False
        # kinematic hardening
        if self.b_k:
            kinematic = True
            assert self.R_0 is not None
            assert self.r_1 is not None
            assert self.r_2 is not None
            if asym:
                assert self.R_0c is not None
                assert self.r_1c is not None
                assert self.r_2c is not None
        else:
            kinematic = False

        #
        # construct argument list
        #

        # these are required and will always be there
        args = ["Steel4", self.uid, self.Fy, self.E0]

        # optional arguments:
        if asym:
            args.extend(['-asym'])
        if kinematic:
            args.extend(['-kin', self.b_k, self.R_0, self.r_1, self.r_2])
            if asym:
                args.extend([self.b_kc, self.R_0c, self.r_1c, self.r_2c])
        if iso:
            args.extend(
                ['-iso', self.b_i, self.rho_i, self.b_l, self.R_i, self.l_yp]
            )
            if asym:
                args.extend([self.b_ic, self.rho_ic, self.b_lc, self.R_ic])
        if ultimate:
            args.extend(['-ult', self.f_u, self.R_u])
            if asym:
                args.extend([self.f_uc, self.R_uc])
        if self.sig_init:
            args.extend(['-init', self.sig_init])
        if self.cycNum:
            args.extend(['-mem', self.cycNum])

        return args


@dataclass
class Bilin(UniaxialMaterial):
    """
    OpenSees Bilin Material
    https://openseespydoc.readthedocs.io/en/latest/src/Bilin.html

    """

    K0: float
    as_Plus: float
    as_Neg: float
    My_Plus: float
    My_Neg: float
    Lamda_S: float
    Lamda_C: float
    Lamda_A: float
    Lamda_K: float
    c_S: float
    c_C: float
    c_A: float
    c_K: float
    theta_p_Plus: float
    theta_p_Neg: float
    theta_pc_Plus: float
    theta_pc_Neg: float
    Res_Pos: float
    Res_Neg: float
    theta_u_Plus: float
    theta_u_Neg: float
    D_Plus: float
    D_Neg: float
    nFactor: float

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees

        """

        return [
            "Bilin",
            self.uid,
            self.K0,
            self.as_Plus,
            self.as_Neg,
            self.My_Plus,
            self.My_Neg,
            self.Lamda_S,
            self.Lamda_C,
            self.Lamda_A,
            self.Lamda_K,
            self.c_S,
            self.c_C,
            self.c_A,
            self.c_K,
            self.theta_p_Plus,
            self.theta_p_Neg,
            self.theta_pc_Plus,
            self.theta_pc_Neg,
            self.Res_Pos,
            self.Res_Neg,
            self.theta_u_Plus,
            self.theta_u_Neg,
            self.D_Plus,
            self.D_Neg,
            self.nFactor,
        ]


@dataclass
class IMKBilin(UniaxialMaterial):
    """
    OpenSees IMKBilin Material
    https://portwooddigital.com/2019/12/08/an-update-of-the-imk-models/

    """

    K0: float
    theta_p_Plus: float
    theta_pc_Plus: float
    theta_u_Plus: float
    My_Plus: float
    as_Plus: float
    Res_Pos: float
    theta_p_Neg: float
    theta_pc_Neg: float
    theta_u_Neg: float
    My_Neg: float
    as_Neg: float
    Res_Neg: float
    Lamda_S: float
    Lamda_C: float
    Lamda_K: float
    c_S: float
    c_C: float
    c_K: float
    D_Plus: float
    D_Neg: float

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees

        """

        return [
            "IMKBilin",
            self.uid,
            self.K0,
            self.theta_p_Plus,
            self.theta_pc_Plus,
            self.theta_u_Plus,
            self.My_Plus,
            self.as_Plus,
            self.Res_Pos,
            self.theta_p_Neg,
            self.theta_pc_Neg,
            self.theta_u_Neg,
            self.My_Neg,
            self.as_Neg,
            self.Res_Neg,
            self.Lamda_S,
            self.Lamda_C,
            self.Lamda_K,
            self.c_S,
            self.c_C,
            self.c_K,
            self.D_Plus,
            self.D_Neg,
        ]


@dataclass
class Pinching4(UniaxialMaterial):
    """
    OpenSees Pinching4 Material
    https://openseespydoc.readthedocs.io/en/latest/src/Pinching4.html

    """

    ePf1: float
    ePf2: float
    ePf3: float
    ePf4: float
    ePd1: float
    ePd2: float
    ePd3: float
    ePd4: float
    eNf1: float
    eNf2: float
    eNf3: float
    eNf4: float
    eNd1: float
    eNd2: float
    eNd3: float
    eNd4: float
    rDispP: float
    fForceP: float
    uForceP: float
    rDispN: float
    fFoceN: float
    uForceN: float
    gK1: float
    gK2: float
    gK3: float
    gK4: float
    gKLim: float
    gD1: float
    gD2: float
    gD3: float
    gD4: float
    gDLim: float
    gF1: float
    gF2: float
    gF3: float
    gF4: float
    gFLim: float
    gE: float
    dmgType: str

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees

        """

        return [
            "Pinching4",
            self.uid,
            self.ePf1,
            self.ePf2,
            self.ePf3,
            self.ePf4,
            self.ePd1,
            self.ePd2,
            self.ePd3,
            self.ePd4,
            self.eNf1,
            self.eNf2,
            self.eNf3,
            self.eNf4,
            self.eNd1,
            self.eNd2,
            self.eNd3,
            self.eNd4,
            self.rDispP,
            self.fForceP,
            self.uForceP,
            self.rDispN,
            self.fFoceN,
            self.uForceN,
            self.gK1,
            self.gK2,
            self.gK3,
            self.gK4,
            self.gKLim,
            self.gD1,
            self.gD2,
            self.gD3,
            self.gD4,
            self.gDLim,
            self.gF1,
            self.gF2,
            self.gF3,
            self.gF4,
            self.gFLim,
            self.gE,
            self.dmgType,
        ]


@dataclass
class Hysteretic(UniaxialMaterial):
    """
    OpenSees Bilin Material
    https://openseespydoc.readthedocs.io/en/latest/src/Bilin.html

    """

    p1: tuple[float, float]
    p2: tuple[float, float]
    p3: tuple[float, float]
    n1: tuple[float, float]
    n2: tuple[float, float]
    n3: tuple[float, float]
    pinchX: float
    pinchY: float
    damage1: float
    damage2: float
    beta: float

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees

        """

        return [
            "Hysteretic",
            self.uid,
            *self.p1,
            *self.p2,
            *self.p3,
            *self.n1,
            *self.n2,
            *self.n3,
            self.pinchX,
            self.pinchY,
            self.damage1,
            self.damage2,
            self.beta,
        ]


@dataclass
class Fatigue(UniaxialMaterial):
    """
    OpenSees Fatigue Material
    https://openseespydoc.readthedocs.io/en/latest/src/Fatigue.html

    """

    predecessor: UniaxialMaterial
    e_mod: float = field(default=0.191)
    var_m: float = field(default=-0.458)
    var_min: float = field(default=-1.0e16)
    var_max: float = field(default=+1.0e16)

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees

        """

        return [
            "Fatigue",
            self.uid,
            self.predecessor.uid,
            "-E0",
            self.e_mod,
            "-m",
            self.var_m,
            "-min",
            self.var_min,
            "-max",
            self.var_max,
        ]


@dataclass
class MaxStrainRange(UniaxialMaterial):
    """
    OpenSees MaxStrainRange Material
    ~not officially added yet~

    """

    predecessor: UniaxialMaterial
    msr_fracture: float
    min_fracture: Optional[float] = field(default=None)
    max_fracture: Optional[float] = field(default=None)
    tangent_ratio: Optional[float] = field(default=None)
    def_coeff: Optional[float] = field(default=None)
    node_tags: Optional[tuple[int, int]] = field(default=None)
    elements_to_remove: Optional[list[int]] = field(default=None)

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees

        """

        args = [
            "MaxStrainRange",
            self.uid,
            self.predecessor.uid,
            self.msr_fracture,
        ]
        if self.min_fracture:
            args.extend(["-min", self.min_fracture])
        if self.max_fracture:
            args.extend(["-max", self.max_fracture])
        if self.tangent_ratio:
            args.extend(["-tangentRatio", self.tangent_ratio])
        if self.def_coeff:
            args.extend(["-defCoeff", self.def_coeff])
        if self.node_tags:
            args.extend(["-nodeTags", *self.node_tags])
        if self.elements_to_remove:
            args.extend(["-eleTag", *self.elements_to_remove])

        return args


@dataclass
class MinMax(UniaxialMaterial):
    """
    OpenSees MinMax Material
    https://openseespydoc.readthedocs.io/en/latest/src/MinMax.html

    """

    predecessor: UniaxialMaterial
    min_strain: float
    max_strain: float

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees

        """

        args = [
            "MinMax",
            self.uid,
            self.predecessor.uid,
            '-min',
            self.min_strain,
            '-max',
            self.max_strain,
        ]

        return args



"""
Model Generator for OpenSees ~ steel design checks
"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from typing import Optional


def smrf_scwb(
    col_sec_properties: dict[str, float],
    col_sec_properties_above: dict[str, float],
    beam_i_sec_properties: dict[str, float],
    col_axial_load: float,
    beam_udl_i: float,
    rbs_proportion_i: float,
    level_height: float,
    bay_length: float,
    beam_j_sec_properties: Optional[dict[str, float]],
    beam_udl_j: Optional[float],
    rbs_proportion_j: Optional[float],
    s_h: float,
    f_y: float,
    ry_coeff: float = 1.15,
    omega_coeff: float = 1.10,
) -> float:
    """
    SMRF strong column weak beam check
    """
    # in3
    zc_below = col_sec_properties["Zx"]
    # in2
    ac_below = col_sec_properties["A"]
    # in3
    zc_above = col_sec_properties_above["Zx"]
    # in2
    ac_above = col_sec_properties_above["A"]
    # lb-in
    mc_below = zc_below * (f_y - col_axial_load / ac_below)
    # lb-in
    mc_above = zc_above * (f_y - col_axial_load / ac_above)
    # lb
    if beam_j_sec_properties:
        max_beam_d = max(
            beam_i_sec_properties["d"], beam_j_sec_properties["d"]
        )
    else:
        max_beam_d = beam_i_sec_properties["d"]
    vc_star = (mc_below + mc_above) / ((level_height - max_beam_d))
    # lb-in
    sigma_mc_star = (mc_below + mc_above) + vc_star * max_beam_d / 2.00
    # in
    c_rbs_i = beam_i_sec_properties["bf"] * (1.0 - rbs_proportion_i) / 2.0
    # in3
    z_rbs_i = beam_i_sec_properties[
        "Zx"
    ] - 2.0 * c_rbs_i * beam_i_sec_properties["tf"] * (
        beam_i_sec_properties["d"] - beam_i_sec_properties["tf"]
    )
    # lb-in
    m_pr_i = ry_coeff * omega_coeff * f_y * z_rbs_i
    # lb
    v_e_i = (2 * m_pr_i) / (bay_length - 2.0 * s_h)
    # lb
    v_g_i = beam_udl_i * (bay_length - 2.0 * s_h) / 2.0
    d_c = col_sec_properties["d"]

    if beam_j_sec_properties:
        assert beam_udl_j is not None
        assert rbs_proportion_j is not None
        c_rbs_j = beam_j_sec_properties["bf"] * (1.0 - rbs_proportion_j) / 2.0
        z_rbs_j = beam_j_sec_properties[
            "Zx"
        ] - 2.0 * c_rbs_j * beam_j_sec_properties["tf"] * (
            beam_j_sec_properties["d"] - beam_j_sec_properties["tf"]
        )
        m_pr_j = ry_coeff * omega_coeff * f_y * z_rbs_j
        v_e_j = (2 * m_pr_j) / (bay_length - 2.0 * s_h)
        v_g_j = beam_udl_j * (bay_length - 2.0 * s_h) / 2.0

        sigm_mb_star = (
            m_pr_i
            + v_e_i * (s_h + d_c / 2.0)
            + v_g_i * (s_h + d_c / 2.0)
            + m_pr_j
            + v_e_j * (s_h + d_c / 2.0)
            - v_g_j * (s_h + d_c / 2.0)
        )
        capacity = sigma_mc_star / sigm_mb_star

    else:
        sigm_mb_star = 1.00 * (
            m_pr_i + v_e_i * (s_h + d_c / 2.0) + v_g_i * (s_h + d_c / 2.0)
        )
        capacity = sigma_mc_star / sigm_mb_star

    return capacity


def smrf_pz_doubler_plate_requirement(
    col_sec_properties: dict[str, float],
    beam_sec_properties: dict[str, float],
    rbs_proportion: float,
    bay_length: float,
    place: str,
    s_h: float,
    f_y: float,
    ry_coeff: float = 1.15,
    omega_coeff: float = 1.10,
) -> float:
    """
    Calculates the required doubler plate thickness
    """
    # in
    c_rbs = beam_sec_properties["bf"] * (1.0 - rbs_proportion) / 2.0
    # in3
    z_rbs = beam_sec_properties["Zx"] - 2.0 * c_rbs * beam_sec_properties[
        "tf"
    ] * (beam_sec_properties["d"] - beam_sec_properties["tf"])
    # lb-in
    m_pr = ry_coeff * omega_coeff * f_y * z_rbs
    # lb
    v_e = 2.0 * m_pr / (bay_length - 2.0 * s_h)
    # lb-in
    m_f = m_pr + v_e * s_h
    # lb
    r_n = (
        0.60
        * f_y
        * col_sec_properties["d"]
        * col_sec_properties["tw"]
        * (
            1.00
            + (
                3.0
                * col_sec_properties["bf"]
                * (col_sec_properties["tf"]) ** 2
            )
            / (
                col_sec_properties["d"]
                * beam_sec_properties["d"]
                * col_sec_properties["tw"]
            )
        )
    )
    if place == "interior":
        r_u = 2 * m_f / (beam_sec_properties["d"] - beam_sec_properties["tf"])
        tdoub = (r_u - r_n) / (0.60 * f_y * col_sec_properties["d"])
        tdoub = max(tdoub, 0.00)
    else:
        r_u = m_f / (beam_sec_properties["d"] - beam_sec_properties["tf"])
        tdoub = (r_u - r_n) / (0.60 * f_y * col_sec_properties["d"])
        tdoub = max(tdoub, 0.00)
    return tdoub


# def steel_W_sec_strength_check(sec_properties, loads):
#     pass



"""
Collection of objects which, as a group, represent some part of a
structure.
"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from dataclasses import dataclass, field
import numpy as np
import numpy.typing as npt
from .ops import element
from . import obj_collections


nparr = npt.NDArray[np.float64]


@dataclass
class ComponentAssembly:
    """
    A component assembly represents some part of a structure and holds
    various lower-level elements such as nodes and beamcolumn
    elements.

    Attributes:
      uid: Unique identifier of the component assembly
      parent_collection: The collection of
        elements to which the component assembly belongs.
      component_purpose: The functionality of the component assembly
      external_nodes: the external nodes to which the
        component assembly is connected.
        these nodes should exist as part of a level.
      internal_nodes: internal nodes that are
        required for the connectivity of the elements of the component
        assembly.
        these nodes only exist as part of the component assembly.
      elements:
        Collection containing the elements that are part of the
        component assembly.

    """

    uid: int
    parent_collection: obj_collections.Collection[int, ComponentAssembly]
    component_purpose: str
    external_nodes: obj_collections.NodeCollection = field(init=False)
    internal_nodes: obj_collections.NodeCollection = field(init=False)
    elements: (
        obj_collections.CollectionWithConnectivity[int, element.Element]
    ) = field(init=False)

    def __post_init__(self):
        self.external_nodes = obj_collections.NodeCollection(self)
        self.internal_nodes = obj_collections.NodeCollection(self)
        self.elements = obj_collections.CollectionWithConnectivity(self)

    def __srepr__(self):
        """
        Short version of repr
        """
        return f"Component assembly, uid: {self.uid}"

    def __repr__(self):
        res = ""
        res += "Component assembly object\n"
        res += f"uid: {self.uid}\n"
        res += f"component_purpose: {self.component_purpose}\n"
        res += "External Nodes\n"
        for node in self.external_nodes.values():
            res += f"  {node.uid}, {node.coords}"
        return res

    def dict_of_elements(self):
        """
        Returns a dictionary of all element objects in the model.
        The keys are the uids of the objects.

        """

        res = {}
        for elm in self.elements.values():
            res[elm.uid] = elm
        return res

    def list_of_elements(self):
        """
        Returns a list of all element objects in the model.

        """

        return list(self.dict_of_elements().values())

    def element_connectivity(self):
        """
        Returns the connectivity of all elements. Elements are
        connected to external nodes. Each component assembly can be
        represented by a tuple of node uids of its connected nodes in
        ascending order. This method returns a dictionary having these
        tuples as keys, and the associated components as values.

        """

        res = {}
        elms = self.list_of_elements()
        for elm in elms:
            uids = [nd.uid for nd in elm.nodes]
            uids.sort()
            uids_tuple = (*uids,)
            res[uids_tuple] = elm
        return res
